#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

#
# Note pickling is used for message serialization.
# This is just for the prototyping & fast PoC, pickling wont be used in the production.
# Instead, protobuf messages will be defined and parsed to avoid malicious pickling.
#

import os
import re
import getpass
import asyncio
import argparse
import binascii
import logging
import requests
from requests.auth import HTTPDigestAuth
import functools
import coloredlogs
import pickle
import sys
import time
import json
import concurrent
import threading
from blessed import Terminal
from cmd2 import Cmd

from . import trace_logger
from . import cli
from . import misc

from monero_glue import agent_lite, agent_misc, trezor_lite
from monero_glue.misc import b58_mnr
from monero_glue.xmr import wallet, monero, crypto, common
from monero_glue.xmr.monero import TsxData
from monero_serialize import xmrtypes

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.WARNING, use_chroot=False)


class TrezorProxy(trezor_lite.TrezorLite):
    """
    Trezor proxy calls to the remote server
    """
    def __init__(self, url=None, *args, **kwargs):
        super().__init__()
        self.url = 'http://127.0.0.1:46123' if url is None else url
        self.endpoint = '%s/api/v1.0' % self.url

    async def ping(self):
        resp = requests.get('%s/ping' % self.endpoint)
        resp.raise_for_status()
        return resp.json()

    async def watch_only(self):
        resp = requests.get('%s/watch_only' % self.endpoint)
        resp.raise_for_status()
        return resp.json()

    async def transfer(self, method, cmd, payload):
        endp = '%s/%s' % (self.endpoint, method)
        req = {'cmd': cmd, 'payload': payload}
        resp = requests.post(endp, json=req)
        resp.raise_for_status()
        return resp.json()

    async def transfer_pickle(self, method, action, *args, **kwargs):
        logger.debug('Action: %s' % action)
        to_pickle = (args, kwargs)
        pickled_data = pickle.dumps(to_pickle)
        payload = binascii.hexlify(pickled_data).decode('utf8')

        resp = await self.transfer(method, action, payload)
        pickle_data = binascii.unhexlify(resp['payload'].encode('utf8'))
        logger.debug('Req size: %s, response size: %s' % (len(pickled_data), len(pickle_data)))

        res = pickle.loads(pickle_data)
        return res

    async def init_transaction(self, tsx_data: TsxData):
        return await self.transfer_pickle('tx_sign', 'init_transaction', tsx_data)

    async def set_tsx_input(self, src_entr):
        return await self.transfer_pickle('tx_sign', 'set_tsx_input', src_entr)

    async def tsx_inputs_permutation(self, permutation):
        return await self.transfer_pickle('tx_sign', 'tsx_inputs_permutation', permutation)

    async def tsx_input_vini(self, *args, **kwargs):
        return await self.transfer_pickle('tx_sign', 'tsx_input_vini', *args, **kwargs)

    async def set_tsx_output1(self, dst_entr, dst_entr_hmac):
        return await self.transfer_pickle('tx_sign', 'set_tsx_output1', dst_entr, dst_entr_hmac)

    async def all_out1_set(self):
        return await self.transfer_pickle('tx_sign', 'all_out1_set')

    async def tsx_mlsag_done(self):
        return await self.transfer_pickle('tx_sign', 'tsx_mlsag_done')

    async def sign_input(self, *args, **kwargs):
        return await self.transfer_pickle('tx_sign', 'sign_input', *args, **kwargs)

    async def tx_sign_final(self, *args, **kwargs):
        return await self.transfer_pickle('tx_sign', 'final', *args, **kwargs)


class WalletRpc(object):
    """
    RPC helper
    """
    def __init__(self, agent, port=None, creds=None):
        self.agent = agent
        self.port = port
        self.creds = creds
        self.url = None
        self.set_addr('127.0.0.1:%s' % port)

    def set_addr(self, addr):
        self.url = 'http://%s/json_rpc' % addr

    def set_creds(self, creds):
        if creds is None or (isinstance(creds, (list, tuple)) and len(creds) == 2):
            self.creds = creds
        elif isinstance(creds, str):
            self.creds = creds.split(':', 1)
        else:
            raise ValueError('Unknown creds type')

    def request(self, method, params=None):
        """
        Request wrapper
        {"jsonrpc":"2.0","id":"0","method":"get_address", "params":{"account_index": 0, "address_index": [0,1,2,3,4,5]}
        :param method:
        :param params:
        :return:
        """
        auth = HTTPDigestAuth(self.creds[0], self.creds[1]) if self.creds else None
        js = {'jsonrpc': '2.0', 'id': '0', 'method': method}
        if params:
            js['params'] = params

        resp = requests.post(self.url, json=js, auth=auth)
        resp.raise_for_status()
        return resp.json()

    def balance(self):
        return self.request('getbalance')

    def height(self):
        return self.request('getheight')

    def get_transfers(self, params=None):
        return self.request('get_transfers', params)

    def rescan_bc(self):
        return self.request('rescan_blockchain')

    def transfer(self, params):
        return self.request('transfer', params)

    def submit_transfer(self, params):
        return self.request('submit_transfer', params)


class HostAgent(cli.BaseCli):
    """
    Host agent wrapper
    """
    prompt = '$> '

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = None

        self.network_type = None
        self.address = None
        self.priv_view = None
        self.pub_view = None
        self.pub_spend = None
        self.network_type = None
        self.wallet_password = b''
        self.wallet_file = None
        self.monero_bin = None
        self.rpc_addr = '127.0.0.1:18081'
        self.rpc_passwd = None
        self.rpc_bind_port = 48084
        self.rpc_running = False

        self.trace_logger = trace_logger.Tracelogger(logger)
        self.loop = asyncio.get_event_loop()
        self.worker_loop = asyncio.new_event_loop()
        self.worker_thread = threading.Thread(target=self.looper, args=(self.worker_loop, ))
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()
        self.wallet_thread = None
        self.terminating = False

        self.trezor_proxy = TrezorProxy()
        self.agent = agent_lite.Agent(self.trezor_proxy)
        self.wallet_proxy = WalletRpc(self, self.rpc_bind_port, self.rpc_passwd)

    def looper(self, loop):
        """
        Main looper
        :param loop:
        :return:
        """
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def submit_coro(self, coro):
        """
        Submits corroutine to the worker loop
        :param fnc:
        :param args:
        :param kwargs:
        :return:
        """
        return asyncio.run_coroutine_threadsafe(coro, self.worker_loop)

    def wait_coro(self, coro):
        """
        Runs corouting, waits for result
        :param fnc:
        :param args:
        :param kwargs:
        :return:
        """
        future = self.submit_coro(coro)
        return future.result()

    #
    # CLI
    #

    def update_intro(self):
        """
        Updates intro text for CLI header - adds version to it.
        :return:
        """
        self.intro = '-'*self.get_term_width() + \
                     '\n    Monero Trezor agent\n' + \
                     '\n' + \
                     '-' * self.get_term_width()

    def update_prompt(self):
        """
        Prompt update
        :return:
        """
        rpc_flag = '' if self.rpc_running else 'R!'
        flags = [rpc_flag]
        flags_str = '|'.join(flags)
        flags_suffix = '|' + flags_str if len(flags_str) > 0 else ''

        self.prompt = '[wallet %s%s]: ' % (self.address[:6].decode('ascii'), flags_suffix)

    #
    # Handlers
    #

    def do_quit(self, line):
        self.terminating = True
        return super().do_quit(line)

    do_q = do_quit
    do_Q = do_quit

    def do_address(self, line):
        print(self.address.decode('ascii'))

    def do_ping(self, line):
        try:
            pres = self.wait_coro(self.trezor_proxy.ping())
            print('OK %s' % pres)

        except Exception as e:
            print('Trezor not connected')
            logger.debug(e)

    def do_balance(self, line):
        res = self.wallet_proxy.balance()
        print('Balance: %.5f' % wallet.conv_disp_amount(res['result']['balance']))
        print('Unlocked Balance: %.5f' % wallet.conv_disp_amount(res['result']['unlocked_balance']))

    def do_height(self, line):
        res = self.wallet_proxy.height()
        print('Height: %s' % res['result']['height'])

    def do_get_transfers(self, line):
        res = self.wallet_proxy.get_transfers({'pool': True, 'in': True, 'out': True})
        print(json.dumps(res, indent=2))

    def do_rescan_bc(self, line):
        res = self.wallet_proxy.rescan_bc()
        print(json.dumps(res, indent=2))

    def do_transfer(self, line):
        if len(line) == 0:
            print('Usage: transfer [<priority>] [<ring_size>] <address> <amount> [<payment_id>]')
        parts = [x for x in line.split(' ') if len(x.strip()) > 0]

        res = misc.parse_transfer_cmd(parts)
        return self.transfer_cmd(res)

    def do_sign(self, line):
        self.wait_coro(self.sign_wrap(line))

    complete_sign = Cmd.path_complete

    #
    # Logic
    #

    async def is_connected(self):
        """
        Returns True if Trezor is connected
        :return:
        """
        try:
            await self.trezor_proxy.ping()
            return True

        except Exception as e:
            return False

    async def load_watchonly(self):
        """
        Loads watch-only credentials from connected Trezor
        :return:
        """
        if not await self.is_connected():
            logger.error('Trezor is not connected')
            raise agent_misc.TrezorNotRunning('Could not load watch-only credentials')

        try:
            print('Loading watch-only credentials from Trezor. Please, confirm the request on Trezor.')
            res = await self.trezor_proxy.watch_only()

            self.priv_view = crypto.b16_to_scalar(res['data']['view_key'].encode('utf8'))
            self.address = res['data']['address'].encode('utf8')
            self.network_type = res['data']['network_type']
            await self.open_with_keys(self.priv_view, self.address)

        except Exception as e:
            raise ValueError(e)

    async def open_account(self):
        """
        Opens the watch only account
        :return:
        """
        creds_passed = self.args.view_key is not None and self.args.address is not None
        account_file_set = self.args.account_file is not None
        account_file_ex = os.path.exists(self.args.account_file) if account_file_set else False
        if creds_passed:
            await self.open_account_passed()
        elif account_file_ex:
            await self.open_account_file(self.args.account_file)
        else:
            await self.check_params(True)
            await self.load_watchonly()

        if self.args.rpc_addr:
            self.rpc_addr = self.args.rpc_addr
        if self.args.watch_wallet:
            self.wallet_file = self.args.watch_wallet
        if self.args.monero_bin:
            self.monero_bin = self.args.monero_bin

        if account_file_set and not account_file_ex:
            await self.check_params(True)
            await self.prompt_password(True)

        # Create watch only wallet file for monero-wallet-rpc
        await self.ensure_watch_only()

        # Write acquired data to the account file
        if account_file_set and not account_file_ex:
            await self.save_account(self.args.account_file)

        print('Public spend key: %s' % binascii.hexlify(crypto.encodepoint(self.pub_spend)).decode('ascii'))
        print('Public view key : %s' % binascii.hexlify(crypto.encodepoint(self.pub_view)).decode('ascii'))
        print('Address:          %s' % self.address.decode('utf8'))
        self.update_intro()
        self.update_prompt()

    async def check_params(self, new_wallet=False):
        """
        All params correctly entered?
        :return:
        """
        if not new_wallet:
            return
        if self.args.watch_wallet is None:
            logger.error('--watch-wallet file is not set. Please specify path where to create the monero watch wallet')
            sys.exit(1)
        if self.args.monero_bin is None:
            logger.error('--monero-bin is not set. Please specify path to the monero binaries')
            sys.exit(1)

    async def prompt_password(self, new_wallet=False):
        """
        Prompts password for a new wallet
        :param new_wallet:
        :return:
        """
        if new_wallet:
            passwd = self.ask_password('Creating a new wallet. Please, enter the password: ', True)
        else:
            passwd = self.ask_password('Please, enter the wallet password: ', False)
        return passwd.encode('utf8')

    async def save_account(self, file):
        """
        Stores account data
        :param file:
        :return:
        """
        with open(file, 'w') as fh:
            data = {
                'view_key': binascii.hexlify(crypto.encodeint(self.priv_view)).decode('ascii'),
                'address': self.address.decode('ascii'),
                'network_type': self.network_type,
                'wallet_password': self.wallet_password.decode('utf8'),
                'rpc_addr': self.rpc_addr,
                'wallet_file': self.args.watch_wallet,
                'monero_bin': self.args.monero_bin,
                'WARNING': 'Agent file is not password encrypted in the PoC',
            }
            json.dump(data, fh, indent=2)

    async def check_existing_wallet_file(self, key_file):
        """
        Checks existing wallet file correctness
        :param key_file:
        :return:
        """
        wl = await wallet.load_keys_file(key_file, self.wallet_password)
        addr = wl['key_data']['m_keys']['m_account_address']
        spend_pub = addr['m_spend_public_key']
        view_pub = addr['m_view_public_key']

        match = spend_pub == crypto.encodepoint(self.pub_spend) and view_pub == crypto.encodepoint(self.pub_view)
        net_ver = monero.net_version(self.network_type, False)
        addr = monero.encode_addr(net_ver, spend_pub, view_pub)
        return addr, match

    async def ensure_watch_only(self):
        """
        Ensures watch only wallet for monero exists
        :return:
        """
        if self.args.watch_wallet is None:
            return

        key_file = '%s.keys' % self.args.watch_wallet
        if os.path.exists(key_file):
            logger.debug('Watch only wallet key file exists: %s' % key_file)
            match, addr = False, None
            try:
                addr, match = await self.check_existing_wallet_file(key_file)
            except Exception as e:
                logger.error('Wallet key file processing exception: %s' % e)

            if not match:
                logger.error('Key file address is not correct: %s' % addr)
                print('Please, move the file so Agent can create correct key file')
                sys.exit(2)
            return

        account_keys = xmrtypes.AccountKeys()
        key_data = wallet.WalletKeyData()

        wallet_data = wallet.WalletKeyFile()
        wallet_data.key_data = key_data
        wallet_data.watch_only = 1
        wallet_data.testnet = self.network_type == monero.NetworkTypes.TESTNET

        key_data.m_creation_timestamp = int(time.time())
        key_data.m_keys = account_keys

        account_keys.m_account_address = xmrtypes.AccountPublicAddress(
            m_spend_public_key=crypto.encodepoint(self.pub_spend),
            m_view_public_key=crypto.encodepoint(self.pub_view),
        )
        account_keys.m_spend_secret_key = crypto.encodeint(0)
        account_keys.m_view_secret_key = crypto.encodeint(self.priv_view)

        await wallet.save_keys_file(key_file, self.wallet_password, wallet_data)
        logger.debug('Watch-only wallet keys generated: %s' % key_file)

    async def open_account_passed(self):
        """
        Loads passed credentials
        :return:
        """
        priv_view = self.args.view_key.encode('ascii')
        self.priv_view = crypto.b16_to_scalar(priv_view)
        self.address = self.args.address.encode('ascii')
        self.network_type = monero.NetworkTypes.TESTNET if self.args.testnet else monero.NetworkTypes.MAINNET
        self.wallet_file = self.args.watch_wallet
        self.monero_bin = self.args.monero_bin
        await self.open_with_keys(self.priv_view, self.address)

    async def open_account_file(self, file):
        """
        Opens account file
        :param file:
        :return:
        """
        with open(file) as fh:
            js = json.load(fh)

        self.wallet_password = await self.prompt_password()

        # Note the agent is not encrypted for PoC - demo.
        self.priv_view = crypto.b16_to_scalar(js['view_key'].encode('utf8'))
        self.address = js['address'].encode('utf8')
        self.wallet_file = js['wallet_file']
        self.monero_bin = js['monero_bin']
        self.network_type = js['network_type']
        self.rpc_addr = js['rpc_addr']

        if self.wallet_password != js['wallet_password'].encode('utf8'):
            raise ValueError('Password didnt match')

        await self.open_with_keys(self.priv_view, self.address)

    async def open_with_keys(self, view_key, address):
        """
        Processess view key private + address
        :param view_key:
        :param address:
        :return:
        """
        self.pub_view = crypto.scalarmult_base(view_key)

        version, pub_spend, pub_view = monero.decode_addr(address)
        self.pub_spend = crypto.decodepoint(pub_spend)

        if not crypto.point_eq(self.pub_view, crypto.decodepoint(pub_view)):
            raise ValueError('Computed view public key does not match the one from address')

    def wallet_rpc_main(self, *args, **kwargs):
        """
        Wallet RPC thread
        :return:
        """
        rpc_cmd = os.path.join(self.monero_bin, 'monero-wallet-rpc')
        if not os.path.exists(rpc_cmd):
            logger.error('Wallet rpc binary not found: %s' % rpc_cmd)
            sys.exit(1)

        self.rpc_passwd = misc.gen_simple_passwd(16)
        self.wallet_proxy.set_creds(['trezor', self.rpc_passwd])

        # TODO: pass via config-file. Passwords visible via proclist. ideally ENV VARS
        args = ['--daemon-address %s' % misc.escape_shell(self.rpc_addr),
                '--wallet-file %s' % misc.escape_shell(self.wallet_file),
                '--password %s' % misc.escape_shell(self.wallet_password),
                '--rpc-login=%s' % ('trezor:%s' % self.rpc_passwd),
                '--rpc-bind-port %s' % int(self.rpc_bind_port),
        ]

        if self.args.testnet or self.network_type == monero.NetworkTypes.TESTNET:
            args.append('--testnet')

        cmd = '%s %s' % (rpc_cmd, ' '.join(args))

        feeder = misc.Feeder()
        p = misc.run(cmd, input=feeder, async=True,
                     stdout=misc.Capture(timeout=1, buffer_size=1),
                     stderr=misc.Capture(timeout=1, buffer_size=1),
                     cwd=os.getcwd(),
                     env=None,
                     shell=True)

        ret_code = 1
        out_acc, err_acc = [], []
        try:
            self.rpc_running = True
            self.update_prompt()
            while len(p.commands) == 0:
                time.sleep(0.15)

            while p.commands[0].returncode is None:
                out, err = p.stdout.read(1), p.stderr.read(1)
                if not common.is_empty(out):
                    out_acc.append(out.decode('utf8'))
                if not common.is_empty(err):
                    err_acc.append(err.decode('utf8'))

                p.commands[0].poll()
                if self.terminating:
                    feeder.feed('quit\n\n')
                    misc.sarge_sigint(p.commands[0])
                    p.close()

                time.sleep(0.01)
            ret_code = p.commands[0].returncode
            out_acc = misc.add_readlines(p.stdout.readlines(), out_acc)
            err_acc = misc.add_readlines(p.stderr.readlines(), err_acc)
            self.rpc_running = False
            self.update_prompt()

            if not self.terminating:
                logger.error('Wallet RPC ended prematurely with code: %s' % ret_code)
                logger.info('Command: %s' % cmd)
                logger.info('Std out: %s' % ''.join(out_acc))
                logger.info('Error out: %s' % ''.join(err_acc))

        except Exception as e:
            logger.error('Exception in wallet RPC command: %s' % e)
            self.trace_logger.log(e)

    def shutdown_rpc(self):
        """
        Waits for rpc shutdown
        :return:
        """
        if self.args.wallet_rpc_addr:  # using already running rpc
            return
        if not self.rpc_running:
            return

        logger.info('Waiting for wallet-RPC to terminate...')
        self.terminating = True
        while self.rpc_running:
            time.sleep(0.1)

    async def wallet_rpc(self):
        """
        Starts wallet RPC server
        :return:
        """
        if self.args.wallet_rpc_addr:  # using existing RPC?
            self.wallet_proxy.set_addr(self.args.wallet_rpc_addr)
            self.wallet_proxy.set_creds(self.args.wallet_rpc_creds)
            self.rpc_running = True
            self.update_prompt()
            return

        self.wallet_thread = threading.Thread(target=self.wallet_rpc_main, args=(None,))
        self.wallet_thread.setDaemon(False)
        self.wallet_thread.start()

    async def entry(self):
        """
        Entry point
        :return:
        """
        if self.args.debug:
            coloredlogs.install(level=logging.DEBUG, use_chroot=False)
        misc.install_sarge_filter()

        await self.open_account()
        await self.wallet_rpc()

        if self.args.sign:
            return await self.sign_wrap(self.args.sign)

        self.update_intro()
        self.cmdloop()
        self.shutdown_rpc()
        logger.info('Terminating')

    #
    # Sign op
    #

    def transfer_cmd(self, parts):
        """
        Transfer logic
        :param parts:
        :return:
        """
        priority, mixin, address, amount, payment_id = parts
        try:
            address_b = address.encode('ascii')
            version, pub_spend_key, pub_view_key = monero.decode_addr(address_b)

        except Exception as e:
            print('Address invalid: %s ' % address)
            return

        print('Sending %s monero to %s' % (amount, address))
        print('Priority: %s, mixin: %s, payment_id: %s'
              % (priority if priority else 'default',
                 mixin if mixin else 'default',
                 payment_id))

        ask_res = self.ask_proceed_quit('Do you confirm (y/n) ? ')
        if ask_res != self.PROCEED_YES:
            return

        params = {
            "destinations":
                [{"amount": int((10**monero.DISPLAY_DECIMAL_POINT) * amount),
                  "address": address},
                 ],
            "account_index": 0,
            "subaddr_indices": [],
            "unlock_time": 0,
            "get_tx_keys": True,
            "do_not_relay": True,
            "get_tx_hex": False,
            "get_tx_metadata": False
        }
        if priority is not None:
            params['priority'] = priority

        if mixin is not None:
            params['mixin'] = mixin

        # Call RPC to prepare unsigned transaction
        self.transfer_params(params)

    def transfer_params(self, params):
        res = self.wallet_proxy.transfer(params)
        result = res['result']
        print('Fee: %s' % wallet.conv_disp_amount(result['fee']))

        ask_res = self.ask_proceed_quit('Do you confirm (y/n) ? ')
        if ask_res != self.PROCEED_YES:
            return

        if 'unsigned_txset' not in result:
            logger.error('Unsigned transaction not found in the response. '
                         'Please make sure you are using compatible monero-wallet-rpc')
            logger.debug(res)
            return

        unsigned = binascii.unhexlify(result['unsigned_txset'])
        self.wait_coro(self.sign_unsigned(unsigned))

    async def sign_unsigned(self, unsigned_txset):
        """
        Signs unsigned txset with the Trezor
        :param unsigned_txset:
        :return:
        """
        res = await self.sign_wrap(fdata=unsigned_txset)
        if isinstance(res, int):
            logger.error('Error')
            return

        print('Transaction has been signed. ')
        ask_res = self.ask_proceed_quit('Do you wish to submit (y/n) ? ')
        if ask_res != self.PROCEED_YES:
            return

        params = {
            'tx_data_hex': binascii.hexlify(res).decode('ascii')
        }

        res = self.wallet_proxy.submit_transfer(params)
        try:
            if len(res['result']['tx_hash_list']) == 0:
                raise ValueError('Transaction submit failed')

            print('SUCCESS: Transaction has been submitted!')

        except Exception as e:
            print('Transaction submit failed')

    async def sign_wrap(self, file=None, fdata=None):
        """
        Sign wrapper
        :param file:
        :param fdata:
        :return:
        """
        if not self.priv_view:
            logger.error('View key not set, cannot sign')
            return -3

        try:
            return await self.sign(file, fdata)

        except agent_misc.TrezorReturnedError as e:
            self.trace_logger.log(e)
            print('Trezor returned an error: %s' % e)
            return 1

        except agent_misc.TrezorNotRunning as e:
            logger.error('Trezor server is not running')
            return 2

    async def sign(self, file=None, fdata=None):
        """
        Performs TX signature
        :param file:
        :param fdata:
        :return:
        """
        try:
            await self.trezor_proxy.ping()
        except Exception as e:
            raise agent_misc.TrezorNotRunning(e)

        if file and not os.path.exists(file):
            raise ValueError('Could not find unsigned transaction file')

        data = fdata
        if data is None:
            with open(file, 'rb') as fh:
                data = fh.read()

        msg = await wallet.load_unsigned_tx(self.priv_view, data)

        # Key image sync
        # key_images = await self.agent.import_outputs(msg.transfers)
        # For now sync only spent key images to the hot wallet.
        key_images = [td.m_key_image for td in msg.transfers]

        txes = []
        pendings = []
        for tx in msg.txes:  # type: xmrtypes.TxConstructionData
            print('Signing transaction with Trezor')
            print('Please check the Trezor and confirm / reject the transaction\n')

            res = await self.agent.sign_transaction_data(tx)
            cdata = self.agent.last_transaction_data()

            # obj = await xmrobj.dump_message(None, res)
            # print(xmrjson.json_dumps(obj, indent=2))

            # Key image sync for spent TXOs
            # Updating only spent.
            for idx in range(len(tx.selected_transfers)):
                idx_mapped = cdata.source_permutation[idx]
                key_images[tx.selected_transfers[idx_mapped]] = res.vin[idx].k_image

            txes.append(await self.agent.serialize_tx(res))
            pending = wallet.construct_pending_tsx(res, tx)
            pendings.append(pending)

            # TODO: store cdata.enc_salt1, cdata.enc_salt2, cdata.enc_keys

        # Key images array has to cover all transfers sent.
        # Watch only wallet does not have key images.
        signed_tx = xmrtypes.SignedTxSet(ptx=pendings, key_images=key_images)
        signed_data = await wallet.dump_signed_tx(self.priv_view, signed_tx)
        with open('signed_monero_tx', 'wb+') as fh:
            fh.write(signed_data)
            logger.info('Signed transaction file: signed_monero_tx')

        print('Key images: %s' % [binascii.hexlify(ff).decode('utf8') for ff in key_images])
        for idx, tx in enumerate(txes):
            fname = 'transaction_%02d' % idx
            with open(fname, 'wb+') as fh:
                fh.write(tx)

            # relay_fname = 'transaction_%02d_relay.sh' % idx
            # hex_ctx = binascii.hexlify(tx).decode('utf8')
            # with open(relay_fname, 'w+') as fh:
            #     fh.write('#!/bin/bash\n')
            #     fh.write('curl -X POST http://%s/sendrawtransaction '
            #              '-d \'{"tx_as_hex":"%s", "do_not_relay":false}\' '
            #              '-H \'Content-Type: application/json\'\n' % (self.args.rpc_addr, hex_ctx))
            #
            # print('Transaction %02d stored to %s, relay script: %s' % (idx, fname, relay_fname))

            # Relay:
            # payload = {'tx_as_hex': hex_ctx, 'do_not_relay': False}
            # resp = requests.post('http://%s/sendrawtransaction' % (self.args.rpc_addr, ), json=payload)
            # print('Relay response: %s' % resp.json())

        # print('Please note that by manual relaying hot wallet key images get out of sync')
        return signed_data

    async def main(self):
        """
        Entry point
        :return:
        """
        parser = argparse.ArgumentParser(description='Trezor Agent')

        parser.add_argument('--address', dest='address',
                            help='Full address')

        parser.add_argument('--view-key', dest='view_key',
                            help='Hex coded private view key')

        parser.add_argument('--account-file', dest='account_file',
                            help='Account file with watch-only creds')

        parser.add_argument('--watch-wallet', dest='watch_wallet',
                            help='Watch-only wallet files')

        parser.add_argument('--monero-bin', dest='monero_bin',
                            help='Directory with monero binaries')

        parser.add_argument('--rpc-addr', dest='rpc_addr', default=None,
                            help='RPC address of full node')

        parser.add_argument('--rpc-wallet', dest='wallet_rpc_addr', default=None,
                            help='Use running monero-wallet-rpc')

        parser.add_argument('--rpc-wallet-creds', dest='wallet_rpc_creds', default=None,
                            help='Running monero-wallet-rpc credentials')

        parser.add_argument('--sign', dest='sign', default=None,
                            help='Sign the unsigned file')

        parser.add_argument('--debug', dest='debug', default=False, action='store_const', const=True,
                            help='Debugging output')

        parser.add_argument('--testnet', dest='testnet', default=False, action='store_const', const=True,
                            help='Testnet')

        args_src = sys.argv
        self.args, unknown = parser.parse_known_args(args=args_src[1:])

        if self.args.rpc_addr:
            if not re.match(r'^\[?([.0-9a-f:]+)\]?(:[0-9]+)?$', self.args.rpc_addr):
                logger.error('Invalid deamon address: %s' % self.args.rpc_addr)
                return -1

        sys.argv = [args_src[0]]
        res = await self.entry()
        sys.argv = args_src
        return res


async def main():
    agent = HostAgent()
    res = await agent.main()
    sys.exit(res)


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())
    # loop.run_forever()
    loop.close()





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: https://github.com/monero-project/mininero
# Author: Dusan Klinec, ph4r05, 2018
# see https://eprint.iacr.org/2015/1098.pdf

import logging
from monero_glue.xmr import crypto, common
from monero_serialize import xmrtypes

logger = logging.getLogger(__name__)


def key_zero_vector(rows):
    """
    Empty key vector
    :param rows:
    :return:
    """
    return [0] * rows


def key_vector(rows):
    """
    Empty key vector
    :param rows:
    :return:
    """
    return [None] * rows


def key_matrix(rows, cols):
    """
    first index is columns (so slightly backward from math)
    :param rows:
    :param cols:
    :return:
    """
    rv = [None] * cols
    for i in range(0, cols):
        rv[i] = key_vector(rows)
    return rv


def hash_key_vector(v):
    """
    Hashes key vector
    :param v:
    :return:
    """
    return [crypto.hash_to_ec(vi) for vi in v]


def scalar_mult_base_vector(v):
    """
    Creates vector of points from scalars
    :param v:
    :return:
    """
    return [crypto.scalarmult_base(a) for a in v]


def key_image_vector(x):
    """
    Takes as input a keyvector, returns the keyimage-vector
    :param x:
    :return:
    """
    return [crypto.scalarmult(crypto.hash_to_ec(crypto.scalarmult_base(xx)), xx) for xx in x]


def scalar_gen_vector(n):
    """
    Generates vector of scalars
    :param n:
    :return:
    """
    return [crypto.random_scalar() for i in range(0, n)]


def scalar_gen_matrix(r, c):
    """
    Generates matrix of scalars
    :param r:
    :param c:
    :return:
    """
    rv = key_matrix(r, c)
    for i in range(0, c):
        rv[i] = scalar_gen_vector(r)
    return rv


def add_keys1(a, b, B):
    """
    aG + bB, G is basepoint
    :param a:
    :param b:
    :param B:
    :return:
    """
    return crypto.point_add(crypto.scalarmult_base(a), crypto.scalarmult(B, b))


def add_keys2(a, A, b, B):
    """
    aA + bB
    :param a:
    :param A:
    :param b:
    :param B:
    :return:
    """
    return crypto.point_add(crypto.scalarmult(A, a), crypto.scalarmult(B, b))


def decode_points(vct):
    """
    Decodes vector of points
    :param vct:
    :return:
    """
    return [crypto.decodepoint(x) for x in vct]


def copy_ct_key(ct):
    """
    Ct key copy
    :param ct:
    :return:
    """
    return xmrtypes.CtKey(mask=ct.mask, dest=ct.dest)


def copy_ct_keys(vct):
    """
    Copy of the CtKey vector
    :param vct:
    :return:
    """
    return [copy_ct_key(x) for x in vct]


def decode_ct_keys_points(vct, copy=False):
    """
    Decodes CtKeys vector as points
    :param vct:
    :param copy:
    :return:
    """
    rvct = copy_ct_keys(vct) if copy else vct
    for i in range(len(rvct)):
        rvct[i].mask = crypto.decodepoint(rvct[i].mask)
        rvct[i].dest = crypto.decodepoint(rvct[i].dest)
    return rvct


def decode_ct_keys_matrix_points(mxt, copy=False):
    """
    Decodes CtKeys matrix as points
    :param vct:
    :param copy:
    :return:
    """
    rmxt = key_matrix(len(mxt), len(mxt[0])) if copy else mxt
    for i in range(len(mxt)):
        cur = decode_ct_keys_points(mxt[i], copy)
        if copy:
            rmxt[i] = cur

    return rmxt


def gen_mlsag(pk, xx, index):
    """
    Multilayered Spontaneous Anonymous Group Signatures (MLSAG signatures)

    These are aka MG signatutes in earlier drafts of the ring ct paper
    c.f. http://eprint.iacr.org/2015/1098 section 2.
    keyImageV just does I[i] = xx[i] * Hash(xx[i] * G) for each i

    Gen creates a signature which proves that for some column in the keymatrix "pk"
       the signer knows a secret key for each row in that column
    Ver verifies that the MG sig was created correctly

    :param pk:
    :param xx:
    :param index:
    :return:
    """
    rows = len(xx)
    cols = len(pk)
    logger.debug("Generating MG sig of size %s x %s" % (rows, cols))
    logger.debug("index is: %s, %s" % (index, pk[index]))
    c = [None] * cols
    alpha = scalar_gen_vector(rows)
    I = key_image_vector(xx)
    L = key_matrix(rows, cols)
    R = key_matrix(rows, cols)
    s = key_matrix(rows, cols)

    m = ''.join(pk[0])
    for i in range(1, cols):
        m = m + ''.join(pk[i])

    L[index] = [crypto.scalarmult_base(aa) for aa in alpha]  # L = aG
    Hi = hash_key_vector(pk[index])
    R[index] = [crypto.scalarmult(Hi[ii], alpha[ii]) for ii in range(0, rows)]  # R = aI
    
    oldi = index
    i = (index + 1) % cols
    c[i] = crypto.cn_fast_hash(m + ''.join(L[oldi]) + ''.join(R[oldi]))

    while i != index:
        s[i] = scalar_gen_vector(rows)
        L[i] = [add_keys1(s[i][j], c[i], pk[i][j]) for j in range(0, rows)]

        Hi = hash_key_vector(pk[i])
        R[i] = [add_keys2(s[i][j], Hi[j], c[i], I[j]) for j in range(0, rows)]
        oldi = i
        i = (i + 1) % cols
        c[i] = crypto.cn_fast_hash(m + ''.join(L[oldi]) + ''.join(R[oldi]))

    s[index] = [crypto.sc_mulsub(alpha[j], c[index], xx[j]) for j in range(0, rows)]  # alpha - c * x
    return I, c[0], s


def ver_mlsag(pk, I, c0, s):
    """
    Verify MLSAG
    :param pk:
    :param I:
    :param c0:
    :param s:
    :return:
    """
    rows = len(pk[0])
    cols = len(pk)
    logger.debug("verifying MG sig of dimensions %s x %s" % (rows, cols))
    c = [None] * (cols + 1)
    c[0] = c0
    L = key_matrix(rows, cols)
    R = key_matrix(rows, cols)
    
    m = ''.join(pk[0])
    for i in range(1, cols):
        m = m + ''.join(pk[i])

    i = 0
    while i < cols:
        L[i] = [add_keys1(s[i][j], c[i], pk[i][j]) for j in range(0, rows)]

        Hi = hash_key_vector(pk[i])
        R[i] = [add_keys2( s[i][j], Hi[j], c[i], I[j]) for j in range(0, rows)]

        oldi = i
        i = i + 1
        c[i] = crypto.cn_fast_hash(m + ''.join(L[oldi]) + ''.join(R[oldi]))

    return c0 == c[cols]


#
# Borromean signatures for range proofs
#


def gen_borromean(x, P1, P2, indices):
    """
    Generates Borromean signature for range proofs.
    :return:
    """
    n = len(P1)
    alpha = key_zero_vector(n)
    s1 = key_zero_vector(n)
    kck = common.get_keccak()  # ee computation

    for ii in range(n):
        alpha[ii] = crypto.random_scalar()
        L = crypto.scalarmult_base(alpha[ii])

        if indices[ii] == 0:
            s1[ii] = crypto.random_scalar()
            c = crypto.hash_to_scalar(crypto.encodepoint(L))
            L = add_keys1(s1[ii], c, P2[ii])
            kck.update(crypto.encodepoint(L))

        else:
            kck.update(crypto.encodepoint(L))

    ee = crypto.sc_reduce32(crypto.decodeint(kck.digest()))
    del kck

    s0 = key_zero_vector(n)

    for jj in range(n):
        if not indices[jj]:
            s0[jj] = crypto.sc_mulsub(alpha[jj], x[jj], ee)
        else:
            s0[jj] = crypto.random_scalar()
            LL = add_keys1(s0[jj], ee, P1[jj])
            cc = crypto.hash_to_scalar(crypto.encodepoint(LL))
            s1[jj] = crypto.sc_mulsub(alpha[jj], x[jj], cc)

    return s0, s1, ee


def ver_borromean(P1, P2, s0, s1, ee):
    """
    Verify range proof signature, Borromean
    (c.f. gmax/andytoshi's paper)
    :param P1:
    :param P2:
    :param s0:
    :param s1:
    :param ee:
    :return:
    """
    n = len(P1)
    Lv1 = key_vector(n)
    for ii in range(n):
        LL = add_keys1(s0[ii], ee, P1[ii])
        chash = crypto.hash_to_scalar(crypto.encodepoint(LL))
        Lv1[ii] = add_keys1(s1[ii], chash, P2[ii])

    kck = common.get_keccak()
    for ii in range(n):
        kck.update(crypto.encodepoint(Lv1[ii]))

    # ee_computed = crypto.hash_to_scalar(crypto.encodepoint(Lv1))
    ee_computed = crypto.sc_reduce32(crypto.decodeint(kck.digest()))

    return crypto.sc_eq(ee_computed, ee)


#
# Optimized versions with incremental hashing,
# Simple and full variants for Monero
#


def hasher_message(message):
    """
    Returns incremental hasher for MLSAG
    :param message:
    :return:
    """
    ctx = common.HashWrapper(common.get_keccak())
    ctx.update(message)
    return ctx


def gen_mlsag_assert(pk, xx, kLRki, mscout, index, dsRows):
    """
    Conditions check for gen_mlsag_ext.
    :param pk:
    :param xx:
    :param kLRki:
    :param mscout:
    :param index:
    :param dsRows:
    :return:
    """
    cols = len(pk)
    if cols <= 1:
        raise ValueError('Cols == 1')
    if index >= cols:
        raise ValueError('Index out of range')

    rows = len(pk[0])
    if rows == 0:
        raise ValueError('Empty pk')

    for i in range(cols):
        if len(pk[i]) != rows:
            raise ValueError('pk is not rectangular')
    if len(xx) != rows:
        raise ValueError('Bad xx size')
    if dsRows > rows:
        raise ValueError('Bad dsRows size')
    if (not kLRki or not mscout) and (kLRki or mscout):
        raise ValueError('Only one of kLRki/mscout is present')
    if kLRki and dsRows != 1:
        raise ValueError('Multisig requires exactly 1 dsRows')
    return rows, cols


def gen_mlsag_rows(message, rv, pk, xx, kLRki, index, dsRows, rows, cols):
    """
    MLSAG computation - the part with secret keys
    :param message:
    :param rv:
    :param pk:
    :param xx:
    :param kLRki:
    :param index:
    :param dsRows:
    :param rows:
    :param cols:
    :return:
    """
    Ip = key_vector(dsRows)
    rv.II = key_vector(dsRows)
    alpha = key_vector(rows)
    aG = key_vector(rows)
    rv.ss = key_matrix(rows, cols)
    aHP = key_vector(dsRows)

    hasher = hasher_message(message)

    for i in range(dsRows):
        hasher.update(crypto.encodepoint(pk[index][i]))
        if kLRki:
            alpha[i] = kLRki.k
            rv.II[i] = kLRki.ki
            hasher.update(crypto.encodepoint(kLRki.L))
            hasher.update(crypto.encodepoint(kLRki.R))

        else:
            Hi = crypto.hash_to_ec(crypto.encodepoint(pk[index][i]))  # originally hashToPoint()
            alpha[i] = crypto.random_scalar()
            aG[i] = crypto.scalarmult_base(alpha[i])
            aHP[i] = crypto.scalarmult(Hi, alpha[i])
            rv.II[i] = crypto.scalarmult(Hi, xx[i])
            hasher.update(crypto.encodepoint(aG[i]))
            hasher.update(crypto.encodepoint(aHP[i]))

        Ip[i] = crypto.precomp(rv.II[i])

    for i in range(dsRows, rows):
        alpha[i] = crypto.random_scalar()
        aG[i] = crypto.scalarmult_base(alpha[i])
        hasher.update(crypto.encodepoint(pk[index][i]))
        hasher.update(crypto.encodepoint(aG[i]))

    c_old = hasher.digest()
    c_old = crypto.sc_reduce32(crypto.decodeint(c_old))
    return c_old, Ip, alpha


def gen_mlsag_ext(message, pk, xx, kLRki, mscout, index, dsRows):
    """
    Multilayered Spontaneous Anonymous Group Signatures (MLSAG signatures)

    :param message:
    :param pk: matrix of points, point form (not encoded)
    :param xx:
    :param kLRki:
    :param mscout:
    :param index:
    :param dsRows:
    :return:
    """
    rows, cols = gen_mlsag_assert(pk, xx, kLRki, mscout, index, dsRows)

    rv = xmrtypes.MgSig()
    c, L, R, Hi = 0, None, None, None

    c_old, Ip, alpha = gen_mlsag_rows(message, rv, pk, xx, kLRki, index, dsRows, rows, cols)

    i = (index + 1) % cols
    if i == 0:
        rv.cc = c_old

    while i != index:
        rv.ss[i] = scalar_gen_vector(rows)
        hasher = hasher_message(message)

        for j in range(dsRows):
            L = add_keys1(rv.ss[i][j], c_old, pk[i][j])
            Hi = crypto.hash_to_ec(crypto.encodepoint(pk[i][j]))  # originally hashToPoint()
            R = add_keys2(rv.ss[i][j], Hi, c_old, Ip[j])
            hasher.update(crypto.encodepoint(pk[i][j]))
            hasher.update(crypto.encodepoint(L))
            hasher.update(crypto.encodepoint(R))

        for j in range(dsRows, rows):
            L = add_keys1(rv.ss[i][j], c_old, pk[i][j])
            hasher.update(crypto.encodepoint(pk[i][j]))
            hasher.update(crypto.encodepoint(L))

        c = crypto.sc_reduce32(crypto.decodeint(hasher.digest()))
        c_old = c
        i = (i + 1) % cols

        if i == 0:
            rv.cc = c_old

    for j in range(rows):
        rv.ss[index][j] = crypto.sc_mulsub(alpha[j], c, xx[j])  # alpha[j] - c * xx[j]; sc_mulsub in original does c-ab

    if mscout:
        mscout(c)

    return rv, c


def ver_mlsag_assert(pk, rv, dsRows):
    """
    Initial params verification for ver_mlsag
    :param pk:
    :param rv:
    :param dsRows:
    :return:
    """
    cols = len(pk)
    if cols < 2:
        raise ValueError('Error! What is c if cols = 1!')

    rows = len(pk[0])
    if rows == 0:
        raise ValueError('Empty pk')

    for i in range(cols):
        if len(pk[i]) != rows:
            raise ValueError('pk is not rectangular')

    if len(rv.II) != dsRows:
        raise ValueError('Bad II size')
    if len(rv.ss) != cols:
        raise ValueError('Bad rv.ss size')
    for i in range(cols):
        if len(rv.ss[i]) != rows:
            raise ValueError('rv.ss is not rectangular')
    if dsRows > rows:
        raise ValueError('Bad dsRows value')

    return rows, cols


def ver_mlsag_ext(message, pk, rv, dsRows):
    """
    Multilayered Spontaneous Anonymous Group Signatures (MLSAG signatures)
    c.f. http://eprint.iacr.org/2015/1098 section 2.
    keyImageV just does I[i] = xx[i] * Hash(xx[i] * G) for each i

    :param message:
    :param pk: matrix of EC points, point form.
    :param rv:
    :param dsRows:
    :return:
    """
    rows, cols = ver_mlsag_assert(pk, rv, dsRows)
    c_old = rv.cc

    Ip = key_vector(dsRows)
    for i in range(dsRows):
        Ip[i] = crypto.precomp(rv.II[i])

    i = 0
    while i < cols:
        c = 0
        hasher = hasher_message(message)
        for j in range(dsRows):
            L = add_keys1(rv.ss[i][j], c_old, pk[i][j])
            Hi = crypto.hash_to_ec(crypto.encodepoint(pk[i][j]))  # originally hashToPoint()
            R = add_keys2(rv.ss[i][j], Hi, c_old, Ip[j])
            hasher.update(crypto.encodepoint(pk[i][j]))
            hasher.update(crypto.encodepoint(L))
            hasher.update(crypto.encodepoint(R))

        for j in range(dsRows, rows):
            L = add_keys1(rv.ss[i][j], c_old, pk[i][j])
            hasher.update(crypto.encodepoint(pk[i][j]))
            hasher.update(crypto.encodepoint(L))

        c = crypto.sc_reduce32(crypto.decodeint(hasher.digest()))
        c_old = c
        i += 1

    c = crypto.sc_sub(c_old, rv.cc)
    return not crypto.sc_isnonzero(c)


def prove_rct_mg(message, pubs, in_sk, out_sk, out_pk, kLRki, mscout, index, txn_fee_key):
    """
    c.f. http://eprint.iacr.org/2015/1098 section 4. definition 10.
    This does the MG sig on the "dest" part of the given key matrix, and
    the last row is the sum of input commitments from that column - sum output commitments
    this shows that sum inputs = sum outputs
    :param message:
    :param pubs: matrix of CtKeys. points are encoded.
    :param in_sk:
    :param out_sk:
    :param out_pk:
    :param kLRki:
    :param mscout:
    :param index:
    :param txn_fee_key:
    :return:
    """
    cols = len(pubs)
    if cols == 0:
        raise ValueError('Empty pubs')
    rows = len(pubs[0])
    if rows == 0:
        raise ValueError('Empty pub row')
    for i in range(cols):
        if len(pubs[i]) != rows:
            raise ValueError('pub is not rectangular')

    if len(in_sk) != rows:
        raise ValueError('Bad inSk size')
    if len(out_sk) != len(out_pk):
        raise ValueError('Bad outsk/putpk size')
    if (not kLRki or not mscout) and (kLRki and mscout):
        raise ValueError('Only one of kLRki/mscout is present')

    sk = key_vector(rows + 1)
    M = key_matrix(rows + 1, cols)
    for i in range(rows + 1):
        sk[i] = 0

    for i in range(cols):
        M[i][rows] = crypto.identity()
        for j in range(rows):
            M[i][j] = crypto.decodepoint(pubs[i][j].dest)
            M[i][rows] = crypto.point_add(M[i][rows], crypto.decodepoint(pubs[i][j].mask))

    sk[rows] = 0
    for j in range(rows):
        sk[j] = in_sk[j].dest
        sk[rows] = crypto.sc_add(sk[rows], in_sk[j].mask)  # add masks in last row

    for i in range(cols):
        for j in range(len(out_pk)):
            M[i][rows] = crypto.point_sub(M[i][rows], crypto.decodepoint(out_pk[j].mask))  # subtract output Ci's in last row

        # Subtract txn fee output in last row
        M[i][rows] = crypto.point_sub(M[i][rows], txn_fee_key)

    for j in range(len(out_pk)):
        sk[rows] = crypto.sc_sub(sk[rows], out_sk[j].mask)  # subtract output masks in last row

    return gen_mlsag_ext(message, M, sk, kLRki, mscout, index, rows)


def prove_rct_mg_simple(message, pubs, in_sk, a, cout, kLRki, mscout, index):
    """
    Simple version for when we assume only
        post rct inputs
        here pubs is a vector of (P, C) length mixin

    :param message:
    :param pubs: vector of CtKeys, public, point values, encoded form. (dest, mask) = (P, C)
    :param in_sk: CtKey, private. (spending private key, input commitment mask (original))
    :param a: mask from the pseudo_output commitment (alpha)
    :param cout: point, decoded. Pseudo output public key.
    :param kLRki:
    :param mscout: lambda accepting c
    :param index:
    :return:
    """
    rows = 1
    cols = len(pubs)
    if cols == 0:
        raise ValueError('Empty pubs')
    if (not kLRki or not mscout) and (kLRki and mscout):
        raise ValueError('Only one of kLRki/mscout is present')

    sk = key_vector(rows + 1)
    M = key_matrix(rows + 1, cols)

    sk[0] = in_sk.dest
    sk[1] = crypto.sc_sub(in_sk.mask, a)

    for i in range(cols):
        M[i][0] = crypto.decodepoint(pubs[i].dest)
        M[i][1] = crypto.point_sub(crypto.decodepoint(pubs[i].mask), cout)

    return gen_mlsag_ext(message, M, sk, kLRki, mscout, index, rows)


def ver_rct_mg(mg, pubs, out_pk, txn_fee_key, message):
    """
    Verifies the above sig is created corretly
    :param mg:
    :param pubs: matrix of EC points, encoded
    :param out_pk:
    :param txn_fee_key:
    :param message:
    :return:
    """
    cols = len(pubs)
    if cols == 0:
        raise ValueError('Empty pubs')
    rows = len(pubs[0])
    if rows == 0:
        raise ValueError('Empty pubs[0]')
    for i in range(cols):
        if len(pubs[i]) != rows:
            raise ValueError('pubs is not rectangular')

    M = key_matrix(rows + 1, cols)
    for i in range(cols):
        M[i][rows] = crypto.identity()

    for j in range(rows):
        for i in range(cols):
            M[i][j] = crypto.decodepoint(pubs[i][j].dest)
            M[i][rows] = crypto.point_add(M[i][rows], crypto.decodepoint(pubs[i][j].mask))  # add Ci in last row

    for i in range(cols):
        for j in range(len(out_pk)):
            M[i][rows] = crypto.point_sub(M[i][rows], crypto.decodepoint(out_pk[j].mask))  # subtract output Ci's in last row

        # subtract txn fee output in last row
        M[i][rows] = crypto.point_sub(M[i][rows], txn_fee_key)

    return ver_mlsag_ext(message, M, mg, rows)


def ver_rct_mg_simple(message, mg, pubs, C):
    """
    Verifies the above sig is created corretly
    :param message:
    :param mg:
    :param pubs: vector of points, encoded
    :param C:
    :return:
    """
    rows = 1
    cols = len(pubs)
    if cols == 0:
        raise ValueError('Empty pubs')

    M = key_matrix(rows + 1, cols)
    for i in range(cols):
        M[i][0] = crypto.decodepoint(pubs[i].dest)
        M[i][1] = crypto.point_sub(crypto.decodepoint(pubs[i].mask), C)

    return ver_mlsag_ext(message, M, mg, rows)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018
# base58 for Monero - working in 8-byte blocks
#
# This gives us eight full-sized blocks and one 5-byte block.
# Eight bytes converts to 11 or less Base58 characters; if a particular block converts to <11 characters,
# the conversion pads it with "1"s (1 is 0 in Base58). Likewise, the final 5-byte block can convert to 7 or
# less Base58 digits; the conversion will ensure the result is 7 digits. Due to the conditional padding,
# the 69-byte string will always convert to 95 Base58 characters (8 * 11 + 7).
# [ https://xmr.llcoins.net/addresstests.html ]

# MoneroPy - A python toolbox for Monero
# Copyright (C) 2016 The MoneroPy Developers.
#
# MoneroPy is released under the BSD 3-Clause license. Use and redistribution of
# this software is subject to the license terms in the LICENSE file found in the
# top-level directory of this distribution.

__alphabet = [ord(s) for s in '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz']
__b58base = 58
__UINT64MAX = 2**64
__encodedBlockSizes = [0, 2, 3, 5, 6, 7, 9, 10, 11]
__fullBlockSize = 8
__fullEncodedBlockSize = 11


def _hexToBin(hex):
    if len(hex) % 2 != 0:
        return "Hex string has invalid length!"
    return [int(hex[i*2:i*2+2], 16) for i in range(len(hex)//2)]


def _binToHex(bin):
    return "".join([("0" + hex(int(bin[i])).split('x')[1])[-2:] for i in range(len(bin))])


def _strToBin(a):
    return [ord(s) for s in a]


def _binToStr(bin):
    return ''.join([chr(bin[i]) for i in range(len(bin))])


def _uint8be_to_64(data):
    l_data = len(data)

    if l_data < 1 or l_data > 8:
        return "Invalid input length"

    res = 0
    switch = 9 - l_data
    for i in range(l_data):
        if switch == 1:
            res = res << 8 | data[i]
        elif switch == 2:
            res = res << 8 | data[i]
        elif switch == 3:
            res = res << 8 | data[i]
        elif switch == 4:
            res = res << 8 | data[i]
        elif switch == 5:
            res = res << 8 | data[i]
        elif switch == 6:
            res = res << 8 | data[i]
        elif switch == 7:
            res = res << 8 | data[i]
        elif switch == 8:
            res = res << 8 | data[i]
        else:
            return "Impossible condition"
    return res


def _uint64_to_8be(num, size):
    res = [0] * size;
    if size < 1 or size > 8:
        return "Invalid input length"

    twopow8 = 2**8
    for i in range(size-1,-1,-1):
        res[i] = num % twopow8
        num = num // twopow8

    return res


def encode_block(data, buf, index):
    l_data = len(data)

    if l_data < 1 or l_data > __fullEncodedBlockSize:
        return "Invalid block length: " + str(l_data)

    num = _uint8be_to_64(data)
    i = __encodedBlockSizes[l_data] - 1

    while num > 0:
        remainder = num % __b58base
        num = num // __b58base
        buf[index+i] = __alphabet[remainder];
        i -= 1

    return buf


def b58encode(data_hex=None, data_bin=None):
    """
    Encode hexadecimal string as base58 (ex: encoding a Monero address).

    :param data_hex:
    :param data_bin:
    :return:
    """
    data = bytearray(data_bin) if data_bin else _hexToBin(data_hex)
    l_data = len(data)

    if l_data == 0:
        return ""

    full_block_count = l_data // __fullBlockSize
    last_block_size = l_data % __fullBlockSize
    res_size = full_block_count * __fullEncodedBlockSize + __encodedBlockSizes[last_block_size]

    res = [0] * res_size
    for i in range(res_size):
        res[i] = __alphabet[0]

    for i in range(full_block_count):
        res = encode_block(data[(i*__fullBlockSize):(i*__fullBlockSize+__fullBlockSize)], res, i * __fullEncodedBlockSize)

    if last_block_size > 0:
        res = encode_block(data[(full_block_count*__fullBlockSize):(full_block_count*__fullBlockSize+last_block_size)], res, full_block_count * __fullEncodedBlockSize)

    return bytes(res)


def decode_block(data, buf, index):
    l_data = len(data)

    if l_data < 1 or l_data > __fullEncodedBlockSize:
        return "Invalid block length: " + l_data

    res_size = __encodedBlockSizes.index(l_data)
    if res_size <= 0:
        return "Invalid block size"

    res_num = 0
    order = 1
    for i in range(l_data-1, -1, -1):
        digit = __alphabet.index(data[i])
        if digit < 0:
            return "Invalid symbol"

        product = order * digit + res_num
        if product > __UINT64MAX:
            return "Overflow"

        res_num = product
        order = order * __b58base

    if res_size < __fullBlockSize and 2**(8 * res_size) <= res_num:
        return "Overflow 2"

    tmp_buf = _uint64_to_8be(res_num, res_size)
    for i in range(len(tmp_buf)):
        buf[i+index] = tmp_buf[i]

    return buf


def b58decode(data_hex=None, data_bin=None):
    """
    Decode a base58 string (ex: a Monero address) into hexidecimal form.

    :param data_hex:
    :param data_bin:
    :return:
    """
    enc = bytearray(data_bin) if data_bin else _strToBin(data_hex)
    l_enc = len(enc)

    if l_enc == 0:
        return ""

    full_block_count = l_enc // __fullEncodedBlockSize
    last_block_size = l_enc % __fullEncodedBlockSize
    last_block_decoded_size = __encodedBlockSizes.index(last_block_size)

    if last_block_decoded_size < 0:
        return "Invalid encoded length"

    data_size = full_block_count * __fullBlockSize + last_block_decoded_size

    data = [0] * data_size
    for i in range(full_block_count):
        data = decode_block(enc[(i*__fullEncodedBlockSize):(i*__fullEncodedBlockSize+__fullEncodedBlockSize)], data, i * __fullBlockSize)

    if last_block_size > 0:
        data = decode_block(enc[(full_block_count*__fullEncodedBlockSize):(full_block_count*__fullEncodedBlockSize+last_block_size)], data, full_block_count * __fullBlockSize)

    return bytes(data)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import os
import unittest
import pkg_resources
import aiounittest
import binascii
import collections

from monero_serialize import xmrserialize, xmrtypes, xmrboost
from monero_glue import trezor_lite, agent_lite
from monero_glue.xmr import wallet, monero, crypto
import zlib


class AgentLiteTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(AgentLiteTest, self).__init__(*args, **kwargs)

    async def test_tx_sign_simple(self):
        """
        Testing tx signature, simple, multiple inputs
        :return:
        """
        unsigned_tx_c = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_uns01.txt'))
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))

        await self.tx_sign_unsigned(unsigned_tx)

    async def test_tx_sign(self):
        """
        Testing tx signature, one input. non-simple RCT
        :return:
        """
        unsigned_tx_c = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_uns02.txt'))
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))
        await self.tx_sign_unsigned(unsigned_tx)

    async def test_tx_sign_sub_dest(self):
        """
        Testing tx signature, one input. non-simple RCT
        :return:
        """
        unsigned_tx_c = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_uns03.txt'))
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))
        await self.tx_sign_unsigned(unsigned_tx)

    async def test_tx_sign_sub_2dest(self):
        """
        Testing tx signature, one input. non-simple RCT
        :return:
        """
        unsigned_tx_c = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_uns04.txt'))
        unsigned_tx = zlib.decompress(binascii.unhexlify(unsigned_tx_c))
        await self.tx_sign_unsigned(unsigned_tx)

    async def test_tx_sign_pending01_boost_full_2dest(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        pending_hex = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_pending01.txt'))
        pending_bin = binascii.unhexlify(pending_hex)
        await self.tx_sign_pending_boost(pending_bin)

    async def test_tx_sign_pending02_boost_full_sub_3dest(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        pending_hex = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_pending02.txt'))
        pending_bin = binascii.unhexlify(pending_hex)
        await self.tx_sign_pending_boost(pending_bin)

    async def test_tx_sign_uns01_boost_full_2dest(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        unsigned_tx_c = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_uns_enc01.txt'))

        creds = self.get_creds()
        unsigned_tx = await wallet.load_unsigned_tx(creds.view_key_private, unsigned_tx_c)
        await self.tx_sign_unsigned_msg(unsigned_tx)

    async def test_tx_sign_uns01_boost_full_2dest_sub(self):
        """
        Testing tx signature, one input. full RCT, boost serialized
        :return:
        """
        unsigned_tx_c = pkg_resources.resource_string(__name__, os.path.join('data', 'tsx_uns_enc02.txt'))

        creds = self.get_creds()
        unsigned_tx = await wallet.load_unsigned_tx(creds.view_key_private, unsigned_tx_c)
        await self.tx_sign_unsigned_msg(unsigned_tx)

    async def tx_sign_unsigned_msg(self, unsigned_tx):
        """
        Signs tx stored in unsigned tx message
        :param unsigned_tx:
        :return:
        """
        tagent = self.init_agent()
        txes = await tagent.sign_unsigned_tx(unsigned_tx)
        return await self.receive(txes[0])

    async def tx_sign_unsigned(self, unsigned_tx):
        """
        Tx sign test with given unsigned transaction data
        :param unsigned_tx:
        :return:
        """
        reader = xmrserialize.MemoryReaderWriter(bytearray(unsigned_tx))
        ar = xmrserialize.Archive(reader, False)
        unsig = xmrtypes.UnsignedTxSet()
        await ar.message(unsig)
        await self.tx_sign_unsigned_msg(unsig)

    async def tx_sign_unsigned_boost(self, unsigned_tx):
        """
        Tx sign test with given unsigned transaction data, serialized by boost -
        unsigned tx produced by watch-only cli wallet.

        :param unsigned_tx:
        :return:
        """
        reader = xmrserialize.MemoryReaderWriter(bytearray(unsigned_tx))
        ar = xmrboost.Archive(reader, False)
        unsig = xmrtypes.UnsignedTxSet()
        await ar.root()
        await ar.message(unsig)
        await self.tx_sign_unsigned_msg(unsig)

    async def tx_sign_pending_boost(self, pending_tx):
        """
        Signs transaction produced by the wallet-rpc, metadata parser, boost
        :param metadata:
        :return:
        """
        reader = xmrserialize.MemoryReaderWriter(bytearray(pending_tx))
        ar = xmrboost.Archive(reader, False)
        pending = xmrtypes.PendingTransaction()
        await ar.root()
        await ar.message(pending)

        tagent = self.init_agent()
        txes = await tagent.sign_tx(pending.construction_data)
        await self.receive(txes[0])

    async def receive(self, tx):
        """
        Test transaction receive with known view/spend keys of destinations.
        :return:
        """
        wallet_creds = [self.get_creds(), self.get_creds_01(), self.get_creds_02()]

        # Unserialize the transaction
        tx_obj = xmrtypes.Transaction()
        reader = xmrserialize.MemoryReaderWriter(bytearray(tx))
        ar1 = xmrserialize.Archive(reader, False)

        await ar1.message(tx_obj, msg_type=xmrtypes.Transaction)
        extras = await monero.parse_extra_fields(tx_obj.extra)
        tx_pub = monero.find_tx_extra_field_by_type(extras, xmrtypes.TxExtraPubKey).pub_key
        additional_pub_keys = monero.find_tx_extra_field_by_type(extras, xmrtypes.TxExtraAdditionalPubKeys)
        num_outs = len(tx_obj.vout)
        num_received = 0

        # Try to receive tsx outputs with each account.
        tx_money_got_in_outs = collections.defaultdict(lambda: 0)
        outs = []

        for idx, creds in enumerate(wallet_creds):
            wallet_subs = {}
            for account in range(0, 5):
                monero.compute_subaddresses(creds, account, range(5), wallet_subs)

            derivation = monero.generate_key_derivation(crypto.decodepoint(tx_pub), creds.view_key_private)
            additional_derivations = []
            if additional_pub_keys and additional_pub_keys.data:
                for x in additional_pub_keys.data:
                    additional_derivations.append(
                        monero.generate_key_derivation(crypto.decodepoint(x), creds.view_key_private))

            for ti, to in enumerate(tx_obj.vout):
                tx_scan_info = monero.check_acc_out_precomp(to, wallet_subs, derivation, additional_derivations, ti)
                if not tx_scan_info.received:
                    continue

                num_received += 1
                tx_scan_info = monero.scan_output(creds, tx_obj, ti, tx_scan_info, tx_money_got_in_outs, outs, False)

                # Check spending private key correctness
                self.assertTrue(crypto.point_eq(crypto.decodepoint(tx_obj.rct_signatures.outPk[ti].mask),
                                                crypto.gen_c(tx_scan_info.mask, tx_scan_info.amount)))

                self.assertTrue(crypto.point_eq(crypto.decodepoint(tx_obj.vout[ti].target.key),
                                                crypto.scalarmult_base(tx_scan_info.in_ephemeral)))

        # All outputs have to be successfully received
        self.assertEqual(num_outs, num_received)

    def get_creds(self):
        """
        Wallet credentials
        :return:
        """
        return monero.AccountCreds.new_wallet(
            priv_view_key=crypto.b16_to_scalar(b'4ce88c168e0f5f8d6524f712d5f8d7d83233b1e7a2a60b5aba5206cc0ea2bc08'),
            priv_spend_key=crypto.b16_to_scalar(b'f2644a3dd97d43e87887e74d1691d52baa0614206ad1b0c239ff4aa3b501750a'),
            network_type=monero.NetworkTypes.TESTNET)

    def get_creds_01(self):
        """
        Wallet 02 credentials
        :return:
        """
        return monero.AccountCreds.new_wallet(
            priv_view_key=crypto.b16_to_scalar(b'42ba20adb337e5eca797565be11c9adb0a8bef8c830bccc2df712535d3b8f608'),
            priv_spend_key=crypto.b16_to_scalar(b'b0ef6bd527b9b23b9ceef70dc8b4cd1ee83ca14541964e764ad23f5151204f0f'),
            network_type=monero.NetworkTypes.TESTNET)

    def get_creds_02(self):
        """
        Wallet 01 credentials
        :return:
        """
        return monero.AccountCreds.new_wallet(
            priv_view_key=crypto.b16_to_scalar(b'9e7aba8ae9ee134e5d5464d9145a4db26793d7411af7d06f20e755cb2a5ad50f'),
            priv_spend_key=crypto.b16_to_scalar(b'283d8bab1aeaee8f8b5aed982fc894c67d3e03db9006e488321c053f5183310d'),
            network_type=monero.NetworkTypes.TESTNET)

    def init_trezor(self):
        """
        Initialize new trezor instance
        :return:
        """
        trez = trezor_lite.TrezorLite()
        trez.creds = self.get_creds()
        return trez

    def init_agent(self):
        """
        Initialize new agent instance
        :return:
        """
        return agent_lite.Agent(self.init_trezor())


if __name__ == "__main__":
    unittest.main()  # pragma: no cover





#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2011 thomasv@gitorious
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# list of words from english.h in bitmonero 
words = [
        "abbey",
        "abducts",
        "ability",
        "ablaze",
        "abnormal",
        "abort",
        "abrasive",
        "absorb",
        "abyss",
        "academy",
        "aces",
        "aching",
        "acidic",
        "acoustic",
        "acquire",
        "across",
        "actress",
        "acumen",
        "adapt",
        "addicted",
        "adept",
        "adhesive",
        "adjust",
        "adopt",
        "adrenalin",
        "adult",
        "adventure",
        "aerial",
        "afar",
        "affair",
        "afield",
        "afloat",
        "afoot",
        "afraid",
        "after",
        "against",
        "agenda",
        "aggravate",
        "agile",
        "aglow",
        "agnostic",
        "agony",
        "agreed",
        "ahead",
        "aided",
        "ailments",
        "aimless",
        "airport",
        "aisle",
        "ajar",
        "akin",
        "alarms",
        "album",
        "alchemy",
        "alerts",
        "algebra",
        "alkaline",
        "alley",
        "almost",
        "aloof",
        "alpine",
        "already",
        "also",
        "altitude",
        "alumni",
        "always",
        "amaze",
        "ambush",
        "amended",
        "amidst",
        "ammo",
        "amnesty",
        "among",
        "amply",
        "amused",
        "anchor",
        "android",
        "anecdote",
        "angled",
        "ankle",
        "annoyed",
        "answers",
        "antics",
        "anvil",
        "anxiety",
        "anybody",
        "apart",
        "apex",
        "aphid",
        "aplomb",
        "apology",
        "apply",
        "apricot",
        "aptitude",
        "aquarium",
        "arbitrary",
        "archer",
        "ardent",
        "arena",
        "argue",
        "arises",
        "army",
        "around",
        "arrow",
        "arsenic",
        "artistic",
        "ascend",
        "ashtray",
        "aside",
        "asked",
        "asleep",
        "aspire",
        "assorted",
        "asylum",
        "athlete",
        "atlas",
        "atom",
        "atrium",
        "attire",
        "auburn",
        "auctions",
        "audio",
        "august",
        "aunt",
        "austere",
        "autumn",
        "avatar",
        "avidly",
        "avoid",
        "awakened",
        "awesome",
        "awful",
        "awkward",
        "awning",
        "awoken",
        "axes",
        "axis",
        "axle",
        "aztec",
        "azure",
        "baby",
        "bacon",
        "badge",
        "baffles",
        "bagpipe",
        "bailed",
        "bakery",
        "balding",
        "bamboo",
        "banjo",
        "baptism",
        "basin",
        "batch",
        "bawled",
        "bays",
        "because",
        "beer",
        "befit",
        "begun",
        "behind",
        "being",
        "below",
        "bemused",
        "benches",
        "berries",
        "bested",
        "betting",
        "bevel",
        "beware",
        "beyond",
        "bias",
        "bicycle",
        "bids",
        "bifocals",
        "biggest",
        "bikini",
        "bimonthly",
        "binocular",
        "biology",
        "biplane",
        "birth",
        "biscuit",
        "bite",
        "biweekly",
        "blender",
        "blip",
        "bluntly",
        "boat",
        "bobsled",
        "bodies",
        "bogeys",
        "boil",
        "boldly",
        "bomb",
        "border",
        "boss",
        "both",
        "bounced",
        "bovine",
        "bowling",
        "boxes",
        "boyfriend",
        "broken",
        "brunt",
        "bubble",
        "buckets",
        "budget",
        "buffet",
        "bugs",
        "building",
        "bulb",
        "bumper",
        "bunch",
        "business",
        "butter",
        "buying",
        "buzzer",
        "bygones",
        "byline",
        "bypass",
        "cabin",
        "cactus",
        "cadets",
        "cafe",
        "cage",
        "cajun",
        "cake",
        "calamity",
        "camp",
        "candy",
        "casket",
        "catch",
        "cause",
        "cavernous",
        "cease",
        "cedar",
        "ceiling",
        "cell",
        "cement",
        "cent",
        "certain",
        "chlorine",
        "chrome",
        "cider",
        "cigar",
        "cinema",
        "circle",
        "cistern",
        "citadel",
        "civilian",
        "claim",
        "click",
        "clue",
        "coal",
        "cobra",
        "cocoa",
        "code",
        "coexist",
        "coffee",
        "cogs",
        "cohesive",
        "coils",
        "colony",
        "comb",
        "cool",
        "copy",
        "corrode",
        "costume",
        "cottage",
        "cousin",
        "cowl",
        "criminal",
        "cube",
        "cucumber",
        "cuddled",
        "cuffs",
        "cuisine",
        "cunning",
        "cupcake",
        "custom",
        "cycling",
        "cylinder",
        "cynical",
        "dabbing",
        "dads",
        "daft",
        "dagger",
        "daily",
        "damp",
        "dangerous",
        "dapper",
        "darted",
        "dash",
        "dating",
        "dauntless",
        "dawn",
        "daytime",
        "dazed",
        "debut",
        "decay",
        "dedicated",
        "deepest",
        "deftly",
        "degrees",
        "dehydrate",
        "deity",
        "dejected",
        "delayed",
        "demonstrate",
        "dented",
        "deodorant",
        "depth",
        "desk",
        "devoid",
        "dewdrop",
        "dexterity",
        "dialect",
        "dice",
        "diet",
        "different",
        "digit",
        "dilute",
        "dime",
        "dinner",
        "diode",
        "diplomat",
        "directed",
        "distance",
        "ditch",
        "divers",
        "dizzy",
        "doctor",
        "dodge",
        "does",
        "dogs",
        "doing",
        "dolphin",
        "domestic",
        "donuts",
        "doorway",
        "dormant",
        "dosage",
        "dotted",
        "double",
        "dove",
        "down",
        "dozen",
        "dreams",
        "drinks",
        "drowning",
        "drunk",
        "drying",
        "dual",
        "dubbed",
        "duckling",
        "dude",
        "duets",
        "duke",
        "dullness",
        "dummy",
        "dunes",
        "duplex",
        "duration",
        "dusted",
        "duties",
        "dwarf",
        "dwelt",
        "dwindling",
        "dying",
        "dynamite",
        "dyslexic",
        "each",
        "eagle",
        "earth",
        "easy",
        "eating",
        "eavesdrop",
        "eccentric",
        "echo",
        "eclipse",
        "economics",
        "ecstatic",
        "eden",
        "edgy",
        "edited",
        "educated",
        "eels",
        "efficient",
        "eggs",
        "egotistic",
        "eight",
        "either",
        "eject",
        "elapse",
        "elbow",
        "eldest",
        "eleven",
        "elite",
        "elope",
        "else",
        "eluded",
        "emails",
        "ember",
        "emerge",
        "emit",
        "emotion",
        "empty",
        "emulate",
        "energy",
        "enforce",
        "enhanced",
        "enigma",
        "enjoy",
        "enlist",
        "enmity",
        "enough",
        "enraged",
        "ensign",
        "entrance",
        "envy",
        "epoxy",
        "equip",
        "erase",
        "erected",
        "erosion",
        "error",
        "eskimos",
        "espionage",
        "essential",
        "estate",
        "etched",
        "eternal",
        "ethics",
        "etiquette",
        "evaluate",
        "evenings",
        "evicted",
        "evolved",
        "examine",
        "excess",
        "exhale",
        "exit",
        "exotic",
        "exquisite",
        "extra",
        "exult",
        "fabrics",
        "factual",
        "fading",
        "fainted",
        "faked",
        "fall",
        "family",
        "fancy",
        "farming",
        "fatal",
        "faulty",
        "fawns",
        "faxed",
        "fazed",
        "feast",
        "february",
        "federal",
        "feel",
        "feline",
        "females",
        "fences",
        "ferry",
        "festival",
        "fetches",
        "fever",
        "fewest",
        "fiat",
        "fibula",
        "fictional",
        "fidget",
        "fierce",
        "fifteen",
        "fight",
        "films",
        "firm",
        "fishing",
        "fitting",
        "five",
        "fixate",
        "fizzle",
        "fleet",
        "flippant",
        "flying",
        "foamy",
        "focus",
        "foes",
        "foggy",
        "foiled",
        "folding",
        "fonts",
        "foolish",
        "fossil",
        "fountain",
        "fowls",
        "foxes",
        "foyer",
        "framed",
        "friendly",
        "frown",
        "fruit",
        "frying",
        "fudge",
        "fuel",
        "fugitive",
        "fully",
        "fuming",
        "fungal",
        "furnished",
        "fuselage",
        "future",
        "fuzzy",
        "gables",
        "gadget",
        "gags",
        "gained",
        "galaxy",
        "gambit",
        "gang",
        "gasp",
        "gather",
        "gauze",
        "gave",
        "gawk",
        "gaze",
        "gearbox",
        "gecko",
        "geek",
        "gels",
        "gemstone",
        "general",
        "geometry",
        "germs",
        "gesture",
        "getting",
        "geyser",
        "ghetto",
        "ghost",
        "giant",
        "giddy",
        "gifts",
        "gigantic",
        "gills",
        "gimmick",
        "ginger",
        "girth",
        "giving",
        "glass",
        "gleeful",
        "glide",
        "gnaw",
        "gnome",
        "goat",
        "goblet",
        "godfather",
        "goes",
        "goggles",
        "going",
        "goldfish",
        "gone",
        "goodbye",
        "gopher",
        "gorilla",
        "gossip",
        "gotten",
        "gourmet",
        "governing",
        "gown",
        "greater",
        "grunt",
        "guarded",
        "guest",
        "guide",
        "gulp",
        "gumball",
        "guru",
        "gusts",
        "gutter",
        "guys",
        "gymnast",
        "gypsy",
        "gyrate",
        "habitat",
        "hacksaw",
        "haggled",
        "hairy",
        "hamburger",
        "happens",
        "hashing",
        "hatchet",
        "haunted",
        "having",
        "hawk",
        "haystack",
        "hazard",
        "hectare",
        "hedgehog",
        "heels",
        "hefty",
        "height",
        "hemlock",
        "hence",
        "heron",
        "hesitate",
        "hexagon",
        "hickory",
        "hiding",
        "highway",
        "hijack",
        "hiker",
        "hills",
        "himself",
        "hinder",
        "hippo",
        "hire",
        "history",
        "hitched",
        "hive",
        "hoax",
        "hobby",
        "hockey",
        "hoisting",
        "hold",
        "honked",
        "hookup",
        "hope",
        "hornet",
        "hospital",
        "hotel",
        "hounded",
        "hover",
        "howls",
        "hubcaps",
        "huddle",
        "huge",
        "hull",
        "humid",
        "hunter",
        "hurried",
        "husband",
        "huts",
        "hybrid",
        "hydrogen",
        "hyper",
        "iceberg",
        "icing",
        "icon",
        "identity",
        "idiom",
        "idled",
        "idols",
        "igloo",
        "ignore",
        "iguana",
        "illness",
        "imagine",
        "imbalance",
        "imitate",
        "impel",
        "inactive",
        "inbound",
        "incur",
        "industrial",
        "inexact",
        "inflamed",
        "ingested",
        "initiate",
        "injury",
        "inkling",
        "inline",
        "inmate",
        "innocent",
        "inorganic",
        "input",
        "inquest",
        "inroads",
        "insult",
        "intended",
        "inundate",
        "invoke",
        "inwardly",
        "ionic",
        "irate",
        "iris",
        "irony",
        "irritate",
        "island",
        "isolated",
        "issued",
        "italics",
        "itches",
        "items",
        "itinerary",
        "itself",
        "ivory",
        "jabbed",
        "jackets",
        "jaded",
        "jagged",
        "jailed",
        "jamming",
        "january",
        "jargon",
        "jaunt",
        "javelin",
        "jaws",
        "jazz",
        "jeans",
        "jeers",
        "jellyfish",
        "jeopardy",
        "jerseys",
        "jester",
        "jetting",
        "jewels",
        "jigsaw",
        "jingle",
        "jittery",
        "jive",
        "jobs",
        "jockey",
        "jogger",
        "joining",
        "joking",
        "jolted",
        "jostle",
        "journal",
        "joyous",
        "jubilee",
        "judge",
        "juggled",
        "juicy",
        "jukebox",
        "july",
        "jump",
        "junk",
        "jury",
        "justice",
        "juvenile",
        "kangaroo",
        "karate",
        "keep",
        "kennel",
        "kept",
        "kernels",
        "kettle",
        "keyboard",
        "kickoff",
        "kidneys",
        "king",
        "kiosk",
        "kisses",
        "kitchens",
        "kiwi",
        "knapsack",
        "knee",
        "knife",
        "knowledge",
        "knuckle",
        "koala",
        "laboratory",
        "ladder",
        "lagoon",
        "lair",
        "lakes",
        "lamb",
        "language",
        "laptop",
        "large",
        "last",
        "later",
        "launching",
        "lava",
        "lawsuit",
        "layout",
        "lazy",
        "lectures",
        "ledge",
        "leech",
        "left",
        "legion",
        "leisure",
        "lemon",
        "lending",
        "leopard",
        "lesson",
        "lettuce",
        "lexicon",
        "liar",
        "library",
        "licks",
        "lids",
        "lied",
        "lifestyle",
        "light",
        "likewise",
        "lilac",
        "limits",
        "linen",
        "lion",
        "lipstick",
        "liquid",
        "listen",
        "lively",
        "loaded",
        "lobster",
        "locker",
        "lodge",
        "lofty",
        "logic",
        "loincloth",
        "long",
        "looking",
        "lopped",
        "lordship",
        "losing",
        "lottery",
        "loudly",
        "love",
        "lower",
        "loyal",
        "lucky",
        "luggage",
        "lukewarm",
        "lullaby",
        "lumber",
        "lunar",
        "lurk",
        "lush",
        "luxury",
        "lymph",
        "lynx",
        "lyrics",
        "macro",
        "madness",
        "magically",
        "mailed",
        "major",
        "makeup",
        "malady",
        "mammal",
        "maps",
        "masterful",
        "match",
        "maul",
        "maverick",
        "maximum",
        "mayor",
        "maze",
        "meant",
        "mechanic",
        "medicate",
        "meeting",
        "megabyte",
        "melting",
        "memoir",
        "menu",
        "merger",
        "mesh",
        "metro",
        "mews",
        "mice",
        "midst",
        "mighty",
        "mime",
        "mirror",
        "misery",
        "mittens",
        "mixture",
        "moat",
        "mobile",
        "mocked",
        "mohawk",
        "moisture",
        "molten",
        "moment",
        "money",
        "moon",
        "mops",
        "morsel",
        "mostly",
        "motherly",
        "mouth",
        "movement",
        "mowing",
        "much",
        "muddy",
        "muffin",
        "mugged",
        "mullet",
        "mumble",
        "mundane",
        "muppet",
        "mural",
        "musical",
        "muzzle",
        "myriad",
        "mystery",
        "myth",
        "nabbing",
        "nagged",
        "nail",
        "names",
        "nanny",
        "napkin",
        "narrate",
        "nasty",
        "natural",
        "nautical",
        "navy",
        "nearby",
        "necklace",
        "needed",
        "negative",
        "neither",
        "neon",
        "nephew",
        "nerves",
        "nestle",
        "network",
        "neutral",
        "never",
        "newt",
        "nexus",
        "nibs",
        "niche",
        "niece",
        "nifty",
        "nightly",
        "nimbly",
        "nineteen",
        "nirvana",
        "nitrogen",
        "nobody",
        "nocturnal",
        "nodes",
        "noises",
        "nomad",
        "noodles",
        "northern",
        "nostril",
        "noted",
        "nouns",
        "novelty",
        "nowhere",
        "nozzle",
        "nuance",
        "nucleus",
        "nudged",
        "nugget",
        "nuisance",
        "null",
        "number",
        "nuns",
        "nurse",
        "nutshell",
        "nylon",
        "oaks",
        "oars",
        "oasis",
        "oatmeal",
        "obedient",
        "object",
        "obliged",
        "obnoxious",
        "observant",
        "obtains",
        "obvious",
        "occur",
        "ocean",
        "october",
        "odds",
        "odometer",
        "offend",
        "often",
        "oilfield",
        "ointment",
        "okay",
        "older",
        "olive",
        "olympics",
        "omega",
        "omission",
        "omnibus",
        "onboard",
        "oncoming",
        "oneself",
        "ongoing",
        "onion",
        "online",
        "onslaught",
        "onto",
        "onward",
        "oozed",
        "opacity",
        "opened",
        "opposite",
        "optical",
        "opus",
        "orange",
        "orbit",
        "orchid",
        "orders",
        "organs",
        "origin",
        "ornament",
        "orphans",
        "oscar",
        "ostrich",
        "otherwise",
        "otter",
        "ouch",
        "ought",
        "ounce",
        "ourselves",
        "oust",
        "outbreak",
        "oval",
        "oven",
        "owed",
        "owls",
        "owner",
        "oxidant",
        "oxygen",
        "oyster",
        "ozone",
        "pact",
        "paddles",
        "pager",
        "pairing",
        "palace",
        "pamphlet",
        "pancakes",
        "paper",
        "paradise",
        "pastry",
        "patio",
        "pause",
        "pavements",
        "pawnshop",
        "payment",
        "peaches",
        "pebbles",
        "peculiar",
        "pedantic",
        "peeled",
        "pegs",
        "pelican",
        "pencil",
        "people",
        "pepper",
        "perfect",
        "pests",
        "petals",
        "phase",
        "pheasants",
        "phone",
        "phrases",
        "physics",
        "piano",
        "picked",
        "pierce",
        "pigment",
        "piloted",
        "pimple",
        "pinched",
        "pioneer",
        "pipeline",
        "pirate",
        "pistons",
        "pitched",
        "pivot",
        "pixels",
        "pizza",
        "playful",
        "pledge",
        "pliers",
        "plotting",
        "plus",
        "plywood",
        "poaching",
        "pockets",
        "podcast",
        "poetry",
        "point",
        "poker",
        "polar",
        "ponies",
        "pool",
        "popular",
        "portents",
        "possible",
        "potato",
        "pouch",
        "poverty",
        "powder",
        "pram",
        "present",
        "pride",
        "problems",
        "pruned",
        "prying",
        "psychic",
        "public",
        "puck",
        "puddle",
        "puffin",
        "pulp",
        "pumpkins",
        "punch",
        "puppy",
        "purged",
        "push",
        "putty",
        "puzzled",
        "pylons",
        "pyramid",
        "python",
        "queen",
        "quick",
        "quote",
        "rabbits",
        "racetrack",
        "radar",
        "rafts",
        "rage",
        "railway",
        "raking",
        "rally",
        "ramped",
        "randomly",
        "rapid",
        "rarest",
        "rash",
        "rated",
        "ravine",
        "rays",
        "razor",
        "react",
        "rebel",
        "recipe",
        "reduce",
        "reef",
        "refer",
        "regular",
        "reheat",
        "reinvest",
        "rejoices",
        "rekindle",
        "relic",
        "remedy",
        "renting",
        "reorder",
        "repent",
        "request",
        "reruns",
        "rest",
        "return",
        "reunion",
        "revamp",
        "rewind",
        "rhino",
        "rhythm",
        "ribbon",
        "richly",
        "ridges",
        "rift",
        "rigid",
        "rims",
        "ringing",
        "riots",
        "ripped",
        "rising",
        "ritual",
        "river",
        "roared",
        "robot",
        "rockets",
        "rodent",
        "rogue",
        "roles",
        "romance",
        "roomy",
        "roped",
        "roster",
        "rotate",
        "rounded",
        "rover",
        "rowboat",
        "royal",
        "ruby",
        "rudely",
        "ruffled",
        "rugged",
        "ruined",
        "ruling",
        "rumble",
        "runway",
        "rural",
        "rustled",
        "ruthless",
        "sabotage",
        "sack",
        "sadness",
        "safety",
        "saga",
        "sailor",
        "sake",
        "salads",
        "sample",
        "sanity",
        "sapling",
        "sarcasm",
        "sash",
        "satin",
        "saucepan",
        "saved",
        "sawmill",
        "saxophone",
        "sayings",
        "scamper",
        "scenic",
        "school",
        "science",
        "scoop",
        "scrub",
        "scuba",
        "seasons",
        "second",
        "sedan",
        "seeded",
        "segments",
        "seismic",
        "selfish",
        "semifinal",
        "sensible",
        "september",
        "sequence",
        "serving",
        "session",
        "setup",
        "seventh",
        "sewage",
        "shackles",
        "shelter",
        "shipped",
        "shocking",
        "shrugged",
        "shuffled",
        "shyness",
        "siblings",
        "sickness",
        "sidekick",
        "sieve",
        "sifting",
        "sighting",
        "silk",
        "simplest",
        "sincerely",
        "sipped",
        "siren",
        "situated",
        "sixteen",
        "sizes",
        "skater",
        "skew",
        "skirting",
        "skulls",
        "skydive",
        "slackens",
        "sleepless",
        "slid",
        "slower",
        "slug",
        "smash",
        "smelting",
        "smidgen",
        "smog",
        "smuggled",
        "snake",
        "sneeze",
        "sniff",
        "snout",
        "snug",
        "soapy",
        "sober",
        "soccer",
        "soda",
        "software",
        "soggy",
        "soil",
        "solved",
        "somewhere",
        "sonic",
        "soothe",
        "soprano",
        "sorry",
        "southern",
        "sovereign",
        "sowed",
        "soya",
        "space",
        "speedy",
        "sphere",
        "spiders",
        "splendid",
        "spout",
        "sprig",
        "spud",
        "spying",
        "square",
        "stacking",
        "stellar",
        "stick",
        "stockpile",
        "strained",
        "stunning",
        "stylishly",
        "subtly",
        "succeed",
        "suddenly",
        "suede",
        "suffice",
        "sugar",
        "suitcase",
        "sulking",
        "summon",
        "sunken",
        "superior",
        "surfer",
        "sushi",
        "suture",
        "swagger",
        "swept",
        "swiftly",
        "sword",
        "swung",
        "syllabus",
        "symptoms",
        "syndrome",
        "syringe",
        "system",
        "taboo",
        "tacit",
        "tadpoles",
        "tagged",
        "tail",
        "taken",
        "talent",
        "tamper",
        "tanks",
        "tapestry",
        "tarnished",
        "tasked",
        "tattoo",
        "taunts",
        "tavern",
        "tawny",
        "taxi",
        "teardrop",
        "technical",
        "tedious",
        "teeming",
        "tell",
        "template",
        "tender",
        "tepid",
        "tequila",
        "terminal",
        "testing",
        "tether",
        "textbook",
        "thaw",
        "theatrics",
        "thirsty",
        "thorn",
        "threaten",
        "thumbs",
        "thwart",
        "ticket",
        "tidy",
        "tiers",
        "tiger",
        "tilt",
        "timber",
        "tinted",
        "tipsy",
        "tirade",
        "tissue",
        "titans",
        "toaster",
        "tobacco",
        "today",
        "toenail",
        "toffee",
        "together",
        "toilet",
        "token",
        "tolerant",
        "tomorrow",
        "tonic",
        "toolbox",
        "topic",
        "torch",
        "tossed",
        "total",
        "touchy",
        "towel",
        "toxic",
        "toyed",
        "trash",
        "trendy",
        "tribal",
        "trolling",
        "truth",
        "trying",
        "tsunami",
        "tubes",
        "tucks",
        "tudor",
        "tuesday",
        "tufts",
        "tugs",
        "tuition",
        "tulips",
        "tumbling",
        "tunnel",
        "turnip",
        "tusks",
        "tutor",
        "tuxedo",
        "twang",
        "tweezers",
        "twice",
        "twofold",
        "tycoon",
        "typist",
        "tyrant",
        "ugly",
        "ulcers",
        "ultimate",
        "umbrella",
        "umpire",
        "unafraid",
        "unbending",
        "uncle",
        "under",
        "uneven",
        "unfit",
        "ungainly",
        "unhappy",
        "union",
        "unjustly",
        "unknown",
        "unlikely",
        "unmask",
        "unnoticed",
        "unopened",
        "unplugs",
        "unquoted",
        "unrest",
        "unsafe",
        "until",
        "unusual",
        "unveil",
        "unwind",
        "unzip",
        "upbeat",
        "upcoming",
        "update",
        "upgrade",
        "uphill",
        "upkeep",
        "upload",
        "upon",
        "upper",
        "upright",
        "upstairs",
        "uptight",
        "upwards",
        "urban",
        "urchins",
        "urgent",
        "usage",
        "useful",
        "usher",
        "using",
        "usual",
        "utensils",
        "utility",
        "utmost",
        "utopia",
        "uttered",
        "vacation",
        "vague",
        "vain",
        "value",
        "vampire",
        "vane",
        "vapidly",
        "vary",
        "vastness",
        "vats",
        "vaults",
        "vector",
        "veered",
        "vegan",
        "vehicle",
        "vein",
        "velvet",
        "venomous",
        "verification",
        "vessel",
        "veteran",
        "vexed",
        "vials",
        "vibrate",
        "victim",
        "video",
        "viewpoint",
        "vigilant",
        "viking",
        "village",
        "vinegar",
        "violin",
        "vipers",
        "virtual",
        "visited",
        "vitals",
        "vivid",
        "vixen",
        "vocal",
        "vogue",
        "voice",
        "volcano",
        "vortex",
        "voted",
        "voucher",
        "vowels",
        "voyage",
        "vulture",
        "wade",
        "waffle",
        "wagtail",
        "waist",
        "waking",
        "wallets",
        "wanted",
        "warped",
        "washing",
        "water",
        "waveform",
        "waxing",
        "wayside",
        "weavers",
        "website",
        "wedge",
        "weekday",
        "weird",
        "welders",
        "went",
        "wept",
        "were",
        "western",
        "wetsuit",
        "whale",
        "when",
        "whipped",
        "whole",
        "wickets",
        "width",
        "wield",
        "wife",
        "wiggle",
        "wildly",
        "winter",
        "wipeout",
        "wiring",
        "wise",
        "withdrawn",
        "wives",
        "wizard",
        "wobbly",
        "woes",
        "woken",
        "wolf",
        "womanly",
        "wonders",
        "woozy",
        "worry",
        "wounded",
        "woven",
        "wrap",
        "wrist",
        "wrong",
        "yacht",
        "yahoo",
        "yanks",
        "yard",
        "yawning",
        "yearbook",
        "yellow",
        "yesterday",
        "yeti",
        "yields",
        "yodel",
        "yoga",
        "younger",
        "yoyo",
        "zapped",
        "zeal",
        "zebra",
        "zero",
        "zesty",
        "zigzags",
        "zinger",
        "zippers",
        "zodiac",
        "zombie",
        "zones",
        "zoom"]


n = len(words)


def mn_len():
    return len(words)


def mn_words():
    return words


# Note about US patent no 5892470: Here each word does not represent a given digit.
# Instead, the digit represented by a word is variable, it depends on the previous word.


def mn_swap_endian_4byte(st):
    # this is from moneromoo's code
    r = st[6:8]+st[4:6]+st[2:4]+st[0:2]
    return r


def mn_encode(message):
    out = []
    n = len(words)
    for i in range(0, len(message), 8):
        message = message[0:i] + mn_swap_endian_4byte(message[i:i+8]) + message[i+8:]

    for i in range(len(message)//8):
        word = message[8*i:8*i+8]
        x = int(word, 16)
        w1 = x % n
        w2 = ((x//n) + w1) % n
        w3 = (((x//n)//n) + w2) % n
        out += [words[w1], words[w2], words[w3]]
    return ' '.join(out)


def mn_decode(wlist):
    out = ''
    n = len(words)
    for i in range(len(wlist)//3):  # note 24 / 3 = 8... 12 / 3 = 4..
        word1, word2, word3 = wlist[3*i:3*i+3]
        w1 = words.index(word1)
        w2 = words.index(word2)
        w3 = words.index(word3)
        x = w1 + n * ((n + w2 - w1) % n) +n * n * ((n + w3 - w2) % n)  # as an int
        b = '%08x' % x  # this is big endian!
        out += mn_swap_endian_4byte(b)

    return out


if __name__ == '__main__':
    import sys
    if len(sys.argv) == 1:
        print('I need arguments: a hex string to encode, or a list of words to decode')
    elif len(sys.argv) == 2:
        print(' '.join(mn_encode(sys.argv[1])))
    else:
        print(mn_decode(sys.argv[1:]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

#
# Note pickling is used for message serialization.
# This is just for the prototyping & fast PoC, pickling wont be used in the production.
# Instead, protobuf messages will be defined and parsed to avoid malicious pickling.
#

import sys
import os
import time
import json
import asyncio
import argparse
import binascii
import logging
import traceback
import threading
import coloredlogs
import pickle
import collections
from blessed import Terminal
from cmd2 import Cmd

import eventlet
from eventlet import wsgi
from flask import Flask, jsonify, request, abort
from wsgiref.simple_server import make_server, WSGIServer
from socketserver import ThreadingMixIn

from . import cli
from monero_poc import misc
from monero_glue import trezor_lite, trezor_iface
from monero_glue.xmr import monero, crypto, common, wallet
from monero_glue.xmr.backend import mnemonic
from monero_glue.misc.bip import bip32

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.WARNING, use_chroot=False)
eventlet.monkey_patch(socket=True)


class TrezorInterface(trezor_iface.TrezorInterface):
    def __init__(self, server=None):
        self.server = server
        self.tsx_waiter = misc.CliPrompt(pre_wait_hook=server.update_prompt)
        self.tsx_data = None
        self.confirmed_result = False

    def get_tx_data(self):
        return self.tsx_data

    @property
    def in_confirmation(self):
        return self.tsx_waiter.in_confirmation

    def confirmation(self, confirmed):
        self.confirmed_result = confirmed
        self.tsx_waiter.confirmation(confirmed)

    async def confirm_transaction(self, tsx_data):
        self.confirmed_result = False
        self.tsx_data = tsx_data
        try:
            self.server.on_confirm_start()
            self.tsx_waiter.wait_confirmation()
            return self.confirmed_result

        finally:
            self.tsx_data = None

    async def transaction_signed(self):
        pass

    async def transaction_error(self, *args, **kwargs):
        logger.error('Transaction error: %s %s' % (args, kwargs))

    async def transaction_finished(self):
        self.server.on_transaction_signed()

    async def transaction_step(self, step, sub_step=None):
        logger.debug('Transaction step: %s, sub step: %s' % (step, sub_step))


class TrezorServer(cli.BaseCli):
    """
    Trezor emulator server
    """
    prompt = '$> '

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.trez = None  # type: trezor_lite.TrezorLite
        self.args = None
        self.network_type = None
        self.creds = None  # type: monero.AccountCreds
        self.port = 46123
        self.trez_iface = TrezorInterface(self)
        self.account_data = None

        self.loop = asyncio.get_event_loop()
        self.running = True
        self.stop_event = threading.Event()
        self.local_data = threading.local()
        self.watch_only_waiter = misc.CliPrompt(pre_wait_hook=self.update_prompt)
        self.ui_lock = threading.Lock()
        self.use_werkzeug = False
        self.thread_rest = None
        self.thread_loop = None
        self.rest_loop = None
        self.worker_loop = None

        self.debug = False
        self.server = None
        self.flask = Flask(__name__)

    #
    # CLI related
    #

    def update_intro(self):
        """
        Updates intro text for CLI header - adds version to it.
        :return:
        """
        self.intro = '-'*self.get_term_width() + \
                     '\n    Trezor server\n'

        if self.creds:
            self.intro += ('\n    Account address: %s ' % self.creds.address.decode('utf8'))
        else:
            self.intro += ('\n    Account not initialized, call new_wallet')

        self.intro += ('\n    Running on: 127.0.0.1:%s ' % self.port) + \
                      '\n' + \
                      '-' * self.get_term_width()

    def update_prompt(self):
        """
        Prompt update
        :return:
        """
        flags = []
        if self.watch_only_waiter.in_confirmation:
            flags.append('W?')
        if self.trez_iface.in_confirmation:
            flags.append('T?')

        flags_str = '|'.join(flags)
        flags_suffix = '|' + flags_str if len(flags_str) > 0 else ''

        addr_str = self.creds.address.decode('ascii')[:6] if self.creds else 'UNINIT'
        self.prompt = '[trezor %s%s]: ' % (addr_str, flags_suffix)

    #
    # Server related
    #

    def shutdown_server(self):
        """
        Shutdown flask server
        :return:
        """
        func = request.environ.get('werkzeug.server.shutdown')
        if func is None:
            raise RuntimeError('Not running with the Werkzeug Server')
        func()

    def terminating(self):
        """
        Set state to terminating
        :return:
        """
        self.running = False
        self.stop_event.set()

    def wait_coro(self, coro):
        """
        Waits on the coroutine, for flask server
        :param coro:
        :return:
        """
        # Running on a different thread is not supported in flask:
        # future = asyncio.run_coroutine_threadsafe(self.on_ping(request=request), self.worker_loop)
        # return future.result()

        # Running on the common rest thread
        # return self.rest_loop.run_until_complete(self.on_ping(request=request))

        # Option: create a new loop as this worker is inside own thread
        loop = asyncio.new_event_loop()
        res = loop.run_until_complete(coro)
        loop.close()
        return res

    def init_rest(self):
        """
        Initializes rest server
        :return:
        """
        @self.flask.route('/api/v1.0/ping', methods=['GET'])
        def keep_alive():
            return self.wait_coro(self.on_ping(request=request))

        @self.flask.route('/api/v1.0/watch_only', methods=['GET'])
        def watch_only():
            return self.wait_coro(self.on_watch_only(request=request))

        @self.flask.route('/api/v1.0/tx_sign', methods=['GET', 'POST'])
        def tx_sign():
            return self.wait_coro(self.on_tx_sign(request=request))

    def wsgi_options(self):
        """
        Returns kwargs for wsgi server
        :return:
        """
        kwargs = dict()
        if False:
            kwargs['minimum_chunk_size'] = 1
        return kwargs

    def serve_werkzeug(self):
        """
        Developer local server, not for production use
        :return:
        """
        r = self.flask.run(debug=self.debug, port=self.port, threaded=True)
        logger.info('Started werkzeug server: %s' % r)

    def serve_eventlet(self):
        """
        Eventlet server, fast async, for production use
        Warning: Eventlet is tricky to work concurrently with classical threading locks

        :return:
        """
        listener = eventlet.listen(('0.0.0.0', self.port))
        logger.info('Starting Eventlet listener %s for Flask %s' % (listener, self.flask))
        wsgi.server(listener, self.flask, **self.wsgi_options())

    #
    # Handlers
    #

    async def on_ping(self, request=None):
        """
        Simple ping
        :param request:
        :return:
        """
        return jsonify({'result': True})

    async def on_watch_only(self, request=None):
        """
        Exports watch only credentials
        :param request:
        :return:
        """
        if not self.creds:
            logger.warning('Agent asks for watch-only credentials, Trezor not initialized')
            return abort(406)

        if self.watch_only_waiter.in_confirmation:
            logger.warning('Agent asks for watch-only credentials concurrently')
            return abort(406)
        
        # Prompt user to confirm.
        self.on_watchonly()
        confirmed = self.watch_only_waiter.wait_confirmation()

        if not confirmed:
            logger.warning('Watch only rejected')
            return abort(403)

        logger.info('Returning watch only credentials...')
        res = {
            'view_key': binascii.hexlify(crypto.encodeint(self.creds.view_key_private)).decode('ascii'),
            'address': self.creds.address.decode('ascii'),
            'network_type': self.network_type,
        }
        return jsonify({'result': True, 'data': res})

    async def on_tx_sign(self, request=None):
        """
        TX sign
        :param request:
        :return:
        """
        js = request.json
        cmd = js['cmd']
        logger.debug('Action: %s' % cmd)
        args, kwargs = self.unpickle_args(js) if 'payload' in js else ([], {})

        if not self.trez:
            logger.warning('Transaction signing request on unitialized Trezor')
            return abort(404)

        if cmd == 'init_transaction':
            res = await self.trez.init_transaction(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'set_tsx_input':
            res = await self.trez.set_tsx_input(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'tsx_inputs_permutation':
            res = await self.trez.tsx_inputs_permutation(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'tsx_input_vini':
            res = await self.trez.tsx_input_vini(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'set_tsx_output1':
            res = await self.trez.set_tsx_output1(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'all_out1_set':
            res = await self.trez.all_out1_set(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'tsx_mlsag_done':
            res = await self.trez.tsx_mlsag_done(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'sign_input':
            res = await self.trez.sign_input(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        elif cmd == 'final':
            res = await self.trez.tx_sign_final(*args, **kwargs)
            return jsonify({'result': True, 'payload': self.pickle_res(res)})

        else:
            return abort(405)

    def unpickle_args(self, js):
        return pickle.loads(binascii.unhexlify(js['payload'].encode('utf8')))

    def pickle_res(self, res):
        return binascii.hexlify(pickle.dumps(res)).decode('utf8')

    #
    # Terminal
    #

    do_q = quit
    do_Q = quit

    def do_new_wallet(self, line):
        self.create_wallet(line)

    def do_address(self, line):
        self.poutput(self.creds.address.decode('ascii'))

    def do_keys(self, line):
        print('Spend key priv:   0x%s' % binascii.hexlify(crypto.encodeint(self.creds.spend_key_private)).decode('ascii'))
        print('View key priv:    0x%s' % binascii.hexlify(crypto.encodeint(self.creds.view_key_private)).decode('ascii'))
        print('')
        print('Spend key pub:    0x%s' % binascii.hexlify(crypto.encodepoint(self.creds.spend_key_public)).decode('ascii'))
        print('View key pub:     0x%s' % binascii.hexlify(crypto.encodepoint(self.creds.view_key_public)).decode('ascii'))

    def on_watchonly(self):
        self.poutput('-' * 80)
        self.poutput('Watch-only request received\nEnter W to confirm/reject\n')

    def on_confirm_start(self):
        self.poutput('-' * 80)
        self.poutput('Transaction confirmation procedure\nEnter T to start\n')

    def on_transaction_signed(self):
        self.poutput('-' * 80)
        self.poutput('Transaction was successfully signed\n')

    def conv_disp_amount(self, amount):
        return wallet.conv_disp_amount(amount)

    def do_T(self, line):
        if not self.trez_iface.in_confirmation:
            self.poutput('No transaction in progress')
            return

        tsx_data = self.trez_iface.get_tx_data()
        self.poutput('Confirming transaction:')
        self.poutput('- ' * 40)
        if tsx_data.payment_id:
            self.poutput('  Payment ID: %s' % binascii.hexlify(tsx_data.payment_id).decode('utf8'))

        self.poutput('  Unlock time: %s' % tsx_data.unlock_time)
        self.poutput('  UTXOs: %s' % tsx_data.num_inputs)
        self.poutput('  Mixin: %s' % tsx_data.mixin)

        chg = tsx_data.change_dts
        addr_chg = monero.public_addr_encode(chg.addr, chg.is_subaddress, self.network_type) if chg else None

        for idx, out in enumerate(tsx_data.outputs):
            addr = monero.public_addr_encode(out.addr, out.is_subaddress, self.network_type)
            if addr != addr_chg:
                self.poutput('  Output %2d: %12.8f to %s, sub: %s'
                             % (idx, self.conv_disp_amount(out.amount), addr.decode('utf8'), out.is_subaddress))
            else:
                self.poutput('  Change:    %12.8f to %s, sub: %s'
                             % (self.conv_disp_amount(out.amount), addr.decode('utf8'), out.is_subaddress))

        self.poutput('  Fee: %.8f' % self.conv_disp_amount(tsx_data.fee))
        self.poutput('  Account: %s' % tsx_data.account)

        self.poutput('- ' * 40)
        result = self.select([(0, 'Confirm the transaction'), (1, 'Reject')], 'Do you confirm the transaction? ')
        self.poutput('\n')
        self.trez_iface.confirmation(result == 0)
        self.update_prompt()

    def do_W(self, line):
        if not self.watch_only_waiter.in_confirmation:
            self.poutput('No prompt')
            return

        result = self.select([(0, 'Confirm'), (1, 'Reject')],
                             'Do you confirm sending watch-only credentials to the client? ')

        self.poutput('\n')
        self.watch_only_waiter.confirmation(result == 0)
        self.update_prompt()

    do_t = do_T
    do_w = do_W

    #
    # Work
    #

    def loop_work(self, loop):
        """
        Looping worker
        :return:
        """
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def rest_thread_work(self):
        """
        Main work method for the server - accepting incoming connections.
        :return:
        """
        logger.info('REST thread started %s %s %s dbg: %s'
                    % (os.getpid(), os.getppid(), threading.current_thread(), self.debug))
        try:
            self.worker_loop = asyncio.new_event_loop()
            self.rest_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.rest_loop)

            worker = threading.Thread(target=self.loop_work, args=(self.worker_loop,))
            worker.setDaemon(True)
            worker.start()

            self.init_rest()
            if self.debug or self.use_werkzeug:
                self.serve_werkzeug()
            else:
                self.serve_eventlet()

            logger.info('Terminating flask: %s' % self.flask)

        except Exception as e:
            logger.error('Exception: %s' % e)
            logger.error(traceback.format_exc())

        self.terminating()
        logger.info('Work loop terminated')

    def rest_server_boot(self):
        """
        Boost the server
        :return:
        """
        self.thread_rest = threading.Thread(target=self.rest_thread_work, args=())
        self.thread_rest.setDaemon(True)
        self.thread_rest.start()

    def create_wallet(self, line):
        """
        Creates a new account
        :return:
        """
        if self.args.account_file:
            if os.path.exists(self.args.account_file):
                logger.error('Wallet file exists, could not overwrite')
                return

        print('Generating new wallet...')
        seed = common.random_bytes(32)

        wl = bip32.Wallet.from_master_secret(seed)
        seed_bip32_words, seed_bip32_words_indices = wl.to_seed_words()
        seed_bip32_b58 = wl.serialize_b58()

        # Generate private keys based on the gen mechanism. Bip44 path + Monero backward compatible
        data = wl.get_child_for_path("m/44'/128'/0'/0/0")
        to_hash = data.chain_code + binascii.unhexlify(data.private_key.get_key())

        # to_hash is initial seed in the Monero sense, recoverable from this seed
        hashed = crypto.cn_fast_hash(to_hash)
        electrum_words = mnemonic.mn_encode(binascii.hexlify(hashed))

        keys = monero.generate_monero_keys(hashed)
        spend_sec, spend_pub, view_sec, view_pub = keys

        print('Seed:             0x%s' % binascii.hexlify(seed).decode('ascii'))
        print('Seed bip39 words: %s' % ' '.join(seed_bip32_words))
        print('Seed bip32 b58:   %s' % seed_bip32_b58)

        print('Seed Monero:      0x%s' % binascii.hexlify(hashed).decode('ascii'))
        print('Seed Monero wrds: %s' % electrum_words)

        print('')
        print('Spend key priv:   0x%s' % binascii.hexlify(crypto.encodeint(spend_sec)).decode('ascii'))
        print('View key priv:    0x%s' % binascii.hexlify(crypto.encodeint(view_sec)).decode('ascii'))
        print('')
        print('Spend key pub:    0x%s' % binascii.hexlify(crypto.encodepoint(spend_pub)).decode('ascii'))
        print('View key pub:     0x%s' % binascii.hexlify(crypto.encodepoint(view_pub)).decode('ascii'))

        self.init_with_keys(spend_sec, view_sec)
        print('')
        print('Address:          %s' % self.creds.address.decode('ascii'))

        self.account_data = collections.OrderedDict()
        self.account_data['seed'] = binascii.hexlify(seed).decode('ascii')
        self.account_data['spend_key'] = binascii.hexlify(crypto.encodeint(spend_sec)).decode('ascii')
        self.account_data['view_key'] = binascii.hexlify(crypto.encodeint(view_sec)).decode('ascii')
        self.account_data['meta'] = collections.OrderedDict([
            ('addr', self.creds.address.decode('ascii')),
            ('bip44_seed', binascii.hexlify(seed).decode('ascii')),
            ('bip32_39_words', ' '.join(seed_bip32_words)),
            ('bip32_b58', seed_bip32_b58),
            ('monero_seed', binascii.hexlify(hashed).decode('ascii')),
            ('monero_words', electrum_words),
        ])

        if self.args.account_file:
            with open(self.args.account_file, 'w+') as fh:
                json.dump(self.account_data, fh, indent=2)
        print('Wallet generated')
        self.update_prompt()

    async def open_account(self):
        """
        Handles account open / management
        :return:
        """
        self.update_intro()
        self.update_prompt()

        priv_spend_key = None
        priv_view_key = None

        if not self.args.account_file and not self.args.spend_key:
            logger.debug('No account file nor spend key. Please generate new account')
            return

        acc_file_exists = False
        if self.args.account_file:
            acc_file_exists = os.path.exists(self.args.account_file)

            if acc_file_exists and self.args.spend_key:
                logger.error('Account file exists, spend key is ignored')
            if acc_file_exists and self.args.view_key:
                logger.error('Account file exists, view key is ignored')
            if acc_file_exists:
                with open(self.args.account_file) as fh:
                    self.account_data = json.load(fh)
                priv_spend_key = crypto.decodeint(binascii.unhexlify(self.account_data['spend_key'].encode('ascii')))
                priv_view_key = crypto.decodeint(binascii.unhexlify(self.account_data['view_key'].encode('ascii')))

        if not acc_file_exists and self.args.spend_key:
            priv_view = self.args.view_key.encode('utf8')
            priv_spend = self.args.spend_key.encode('utf8')
            priv_view_key = crypto.b16_to_scalar(priv_view)
            priv_spend_key = crypto.b16_to_scalar(priv_spend)

            self.account_data = {'spend_key': priv_spend_key, 'view_key': priv_view_key}
            with open(self.args.account_file, 'w') as fh:
                json.dump(self.account_data, fh, indent=2)

        if priv_spend_key and priv_view_key:
            self.init_with_keys(priv_spend_key, priv_view_key)

    def init_with_keys(self, priv_spend_key, priv_view_key):
        """
        Initializes Trezor classes with the private keys
        :return:
        """
        self.creds = monero.AccountCreds.new_wallet(priv_view_key, priv_spend_key, self.network_type)
        self.update_intro()

        self.trez = trezor_lite.TrezorLite()
        self.trez.creds = self.creds
        self.trez.iface = self.trez_iface
        self.update_prompt()

    async def entry(self):
        """
        pass
        :return:
        """
        if self.args.debug:
            coloredlogs.install(level=logging.DEBUG, use_chroot=False)

        self.network_type = monero.NetworkTypes.MAINNET
        if self.args.testnet:
            self.network_type = monero.NetworkTypes.TESTNET
        elif self.args.stagenet:
            self.network_type = monero.NetworkTypes.STAGENET

        await self.open_account()

        logging.info('Starting rest server...')
        self.rest_server_boot()
        time.sleep(1)

        self.cmdloop()
        self.terminating()

    async def main(self):
        """
        Entry point
        :return:
        """
        parser = argparse.ArgumentParser(description='Trezor server')

        parser.add_argument('--testnet', dest='testnet', default=False, action='store_const', const=True,
                            help='Testnet')

        parser.add_argument('--stagenet', dest='stagenet', default=False, action='store_const', const=True,
                            help='Stagenet')

        parser.add_argument('--debug', dest='debug', default=False, action='store_const', const=True,
                            help='Debug')

        parser.add_argument('--account-file', dest='account_file',
                            help='Trezor account file to use / open')

        parser.add_argument('--view-key', dest='view_key',
                            help='Hex coded private view key')

        parser.add_argument('--spend-key', dest='spend_key',
                            help='Hex coded private spend key')

        args_src = sys.argv
        self.args, unknown = parser.parse_known_args(args=args_src[1:])

        sys.argv = [args_src[0]]
        await self.entry()
        sys.argv = args_src


async def main():
    agent = TrezorServer()
    await agent.main()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())
    # loop.run_forever()
    loop.close()





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

from monero_glue.xmr import common, crypto


class TrezorError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])


class TrezorSecurityError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorInvalidStateError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorTxPrefixHashNotMatchingError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def compute_tx_key(spend_key_private, tx_prefix_hash, salt=None, rand_mult=None):
    """

    :param spend_key_private:
    :param tx_prefix_hash:
    :param salt:
    :param rand_mult:
    :return:
    """
    if not salt:
        salt = common.random_bytes(32)

    if not rand_mult:
        rand_mult_num = crypto.random_scalar()
        rand_mult = crypto.encodeint(rand_mult_num)
    else:
        rand_mult_num = crypto.decodeint(rand_mult)

    rand_inp = crypto.sc_add(spend_key_private, rand_mult_num)
    passwd = common.keccak_2hash(crypto.encodeint(rand_inp) + tx_prefix_hash)
    tx_key = common.pbkdf2(passwd, salt, count=100)
    return tx_key, salt, rand_mult








#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import argparse
import binascii
from monero_glue.xmr import monero, crypto


def main():
    """
    Entry point
    :return:
    """
    parser = argparse.ArgumentParser(description='Generate non-deterministic wallet credentials')

    parser.add_argument('--testnet', dest='testnet', default=False, action='store_const', const=True,
                        help='Testnet')

    parser.add_argument('--stagenet', dest='stagenet', default=False, action='store_const', const=True,
                        help='Testnet')

    args = parser.parse_args()

    network_type = monero.NetworkTypes.MAINNET
    if args.testnet:
        network_type = monero.NetworkTypes.TESTNET
    elif args.stagenet:
        network_type = monero.NetworkTypes.STAGENET

    priv_view = crypto.random_scalar()
    priv_spend = crypto.random_scalar()
    w = monero.AccountCreds.new_wallet(priv_view, priv_spend, network_type=network_type)

    print('Address: %s' % w.address.decode('utf8'))
    print('Private view key:  %s' % binascii.hexlify(crypto.encodeint(priv_view)).decode('utf8'))
    print('Private spend key: %s' % binascii.hexlify(crypto.encodeint(priv_spend)).decode('utf8'))


if __name__ == '__main__':
    main()





from functools import wraps
import hashlib
from hashlib import sha256
import re

import six

if six.PY3:
    long = int


def ensure_bytes(data):
    if not isinstance(data, six.binary_type):
        return data.encode('utf-8')
    return data


def ensure_str(data):
    if isinstance(data, six.binary_type):
        return data.decode('utf-8')
    elif not isinstance(data, six.string_types):
        raise ValueError("Invalid value for string")
    return data


def chr_py2(num):
    """Ensures that python3's chr behavior matches python2."""
    if six.PY3:
        return bytes([num])
    return chr(num)


def hash160(data):
    """Return ripemd160(sha256(data))"""
    rh = hashlib.new('ripemd160', sha256(data).digest())
    return rh.digest()


def is_hex_string(string):
    """Check if the string is only composed of hex characters."""
    pattern = re.compile(r'[A-Fa-f0-9]+')
    if isinstance(string, six.binary_type):
        string = str(string)
    return pattern.match(string) is not None


def long_to_hex(l, size):
    """Encode a long value as a hex string, 0-padding to size.
    Note that size is the size of the resulting hex string. So, for a 32Byte
    long size should be 64 (two hex characters per byte"."""
    f_str = "{0:0%sx}" % size
    return ensure_bytes(f_str.format(l).lower())


def long_or_int(val, *args):
    return long(val, *args)


def memoize(f):
    """Memoization decorator for a function taking one or more arguments."""
    def _c(*args, **kwargs):
        if not hasattr(f, 'cache'):
            f.cache = dict()
        key = (args, tuple(kwargs))
        if key not in f.cache:
            f.cache[key] = f(*args, **kwargs)
        return f.cache[key]
    return wraps(f)(_c)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: https://github.com/monero-project/mininero
# Author: Dusan Klinec, ph4r05, 2018

import logging

from monero_serialize import xmrtypes
from monero_glue.xmr import mlsag2, crypto, common, asnl

logger = logging.getLogger(__name__)
ATOMS = 64

    
def d2b(n, digits):
    b = [0] * digits
    i = 0
    while n:
        b[i] = n & 1
        i = i + 1
        n >>= 1
    return b 


def sum_Ci(Cis):
    """
    Sums points
    :param Cis:
    :return:
    """
    CSum = crypto.identity()
    for i in Cis:
        CSum = crypto.point_add(CSum, i)
    return CSum


def prove_range(amount, last_mask=None, use_asnl=False, mem_opt=True):
    """
    Range proof generator.
    :param amount:
    :param last_mask:
    :param use_asnl:
    :param mem_opt: memory optimized
    :return:
    """
    if use_asnl and mem_opt:
        raise ValueError('ASNL not in memory optimized variant')

    if use_asnl and not mem_opt:
        return prove_range_orig(amount, last_mask=last_mask, use_asnl=True)
    else:
        return prove_range_mem(amount, last_mask=last_mask)


def prove_range_orig(amount, last_mask=None, use_asnl=False):
    """
    Gives C, and mask such that \sumCi = C
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1

    Ci is a commitment to either 0 or 2^i, i=0,...,63
    thus this proves that "amount" is in [0, 2^ATOMS]
    mask is a such that C = aG + bH, and b = amount
    :param amount:
    :param last_mask: ai[ATOMS-1] will be computed as \sum_{i=0}^{ATOMS-2} a_i - last_mask
    :param use_asnl: use ASNL, used before Borromean
    :return: sumCi, mask, RangeSig.
        sumCi is Pedersen commitment on the amount value. sumCi = aG + amount*H
        mask is "a" from the Pedersent commitment above.
    """
    bb = d2b(amount, ATOMS)  # gives binary form of bb in "digits" binary digits
    logger.info("amount, amount in binary %s %s" % (amount, bb))
    ai = [None] * len(bb)
    Ci = [None] * len(bb)
    CiH = [None] * len(bb)  # this is like Ci - 2^i H
    H2 = crypto.gen_Hpow(ATOMS)
    a = 0
    for i in range(0, ATOMS):
        ai[i] = crypto.random_scalar()
        if last_mask is not None and i == ATOMS - 1:
            ai[i] = crypto.sc_sub(last_mask, a)

        a = crypto.sc_add(a, ai[i])  # creating the total mask since you have to pass this to receiver...
        if bb[i] == 0:
            Ci[i] = crypto.scalarmult_base(ai[i])
        if bb[i] == 1:
            Ci[i] = crypto.point_add(crypto.scalarmult_base(ai[i]), H2[i])
        CiH[i] = crypto.point_sub(Ci[i], H2[i])

    A = xmrtypes.BoroSig()

    if use_asnl:
        A.s0, A.s1, A.ee = asnl.gen_asnl(ai, Ci, CiH, bb)
    else:
        A.s0, A.s1, A.ee = mlsag2.gen_borromean(ai, Ci, CiH, bb)
    
    R = xmrtypes.RangeSig()
    R.asig = A
    R.Ci = Ci

    C = sum_Ci(Ci)
    return C, a, R


def prove_range_mem(amount, last_mask=None):
    """
    Memory optimized range proof.
    Gives C, and mask such that \sumCi = C
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1

    Ci is a commitment to either 0 or 2^i, i=0,...,63
    thus this proves that "amount" is in [0, 2^ATOMS]
    mask is a such that C = aG + bH, and b = amount
    :param amount:
    :param last_mask: ai[ATOMS-1] will be computed as \sum_{i=0}^{ATOMS-2} a_i - last_mask
    :param use_asnl: use ASNL, used before Borromean
    :return: sumCi, mask, RangeSig.
        sumCi is Pedersen commitment on the amount value. sumCi = aG + amount*H
        mask is "a" from the Pedersent commitment above.
    """
    n = ATOMS
    bb = d2b(amount, n)  # gives binary form of bb in "digits" binary digits
    ai = [None] * len(bb)
    Ci = [None] * len(bb)
    a = 0

    C = crypto.identity()
    alpha = mlsag2.key_zero_vector(n)
    s1 = mlsag2.key_zero_vector(n)
    c_H = crypto.gen_H()
    kck = common.get_keccak()  # ee computation

    # First pass, generates: ai, alpha, Ci, ee, s1
    for ii in range(n):
        ai[ii] = crypto.random_scalar()
        if last_mask is not None and ii == ATOMS - 1:
            ai[ii] = crypto.sc_sub(last_mask, a)

        a = crypto.sc_add(a, ai[ii])  # creating the total mask since you have to pass this to receiver...

        alpha[ii] = crypto.random_scalar()
        L = crypto.scalarmult_base(alpha[ii])

        if bb[ii] == 0:
            Ci[ii] = crypto.scalarmult_base(ai[ii])
        else:
            Ci[ii] = crypto.point_add(crypto.scalarmult_base(ai[ii]), c_H)
        C = crypto.point_add(C, Ci[ii])

        if bb[ii] == 0:
            s1[ii] = crypto.random_scalar()
            c = crypto.hash_to_scalar(crypto.encodepoint(L))
            L = mlsag2.add_keys1(s1[ii], c, crypto.point_sub(Ci[ii], c_H))
            kck.update(crypto.encodepoint(L))

        else:
            kck.update(crypto.encodepoint(L))

        c_H = crypto.scalarmult(c_H, 2)

    # Compute ee, memory cleanup
    ee = crypto.sc_reduce32(crypto.decodeint(kck.digest()))
    del kck

    # Second phase computes: s0, s1
    c_H = crypto.gen_H()
    s0 = mlsag2.key_zero_vector(n)

    for jj in range(n):
        if not bb[jj]:
            s0[jj] = crypto.sc_mulsub(alpha[jj], ai[jj], ee)

        else:
            s0[jj] = crypto.random_scalar()
            LL = mlsag2.add_keys1(s0[jj], ee, Ci[jj])
            cc = crypto.hash_to_scalar(crypto.encodepoint(LL))
            s1[jj] = crypto.sc_mulsub(alpha[jj], ai[jj], cc)
        c_H = crypto.scalarmult(c_H, 2)

    A = xmrtypes.BoroSig()
    A.s0, A.s1, A.ee = s0, s1, ee

    R = xmrtypes.RangeSig()
    R.asig = A
    R.Ci = Ci

    return C, a, R


def ver_range(Ci, ags, use_asnl=False):
    """
    Verifies that \sum Ci = C and that each Ci is a commitment to 0 or 2^i
    :param Ci:
    :param ags:
    :param use_asnl: use ASNL, used before Borromean
    :return:
    """
    n = ATOMS
    CiH = [None] * n
    C_tmp = crypto.identity()
    H2 = crypto.gen_Hpow(ATOMS)
    for i in range(0, n):
        CiH[i] = crypto.point_sub(ags.Ci[i], H2[i])
        C_tmp = crypto.point_add(C_tmp, ags.Ci[i])

    if not crypto.point_eq(C_tmp, Ci):
        return 0

    if use_asnl:
        return asnl.ver_asnl(ags.Ci, CiH, ags.asig.s0, ags.asig.s1, ags.asig.ee)
    else:
        return mlsag2.ver_borromean(ags.Ci, CiH, ags.asig.s0, ags.asig.s1, ags.asig.ee)


# Ring-ct MG sigs
# Prove:
#   c.f. http:#eprint.iacr.org/2015/1098 section 4. definition 10. 
#   This does the MG sig on the "dest" part of the given key matrix, and 
#   the last row is the sum of input commitments from that column - sum output commitments
#   this shows that sum inputs = sum outputs
# Ver:
#   verifies the above sig is created corretly


def ecdh_encode(unmasked, receiver_pk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    :param unmasked:
    :param receiver_pk:
    :param derivation:
    :return:
    """
    rv = xmrtypes.EcdhTuple()
    if derivation is None:
        esk = crypto.random_scalar()
        rv.senderPk = crypto.scalarmult_base(esk)
        derivation = crypto.encodepoint(crypto.scalarmult(receiver_pk, esk))

    sharedSec1 = crypto.hash_to_scalar(derivation)
    sharedSec2 = crypto.hash_to_scalar(crypto.encodeint(sharedSec1))

    rv.mask = crypto.sc_add(unmasked.mask, sharedSec1)
    rv.amount = crypto.sc_add(unmasked.amount, sharedSec2)
    return rv

    
def ecdh_decode(masked, receiver_sk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    :param masked:
    :param receiver_sk:
    :param derivation:
    :return:
    """
    rv = xmrtypes.EcdhTuple()

    if derivation is None:
        derivation = crypto.scalarmult(masked.senderPk, receiver_sk)

    sharedSec1 = crypto.hash_to_scalar(derivation)
    sharedSec2 = crypto.hash_to_scalar(crypto.encodeint(sharedSec1))

    rv.mask = crypto.sc_sub(masked.mask, sharedSec1)
    rv.amount = crypto.sc_sub(masked.amount, sharedSec2)
    return rv


#
# RingCT protocol
#


def ver_rct(rv):
    """
    Verifies that all signatures (rangeProogs, MG sig, sum inputs = outputs) are correct

    Inputs:
        - rangesigs is a list of one rangeproof for each output
        - MG is the mgsig [ss, cc, II]
        - mixRing is a ctkeyMatrix
        - ecdhInfo is a list of masks / amounts for each output
        - outPk is a vector of ctkeys (since we have computed the commitment for each amount)
    :param rv: [rangesigs, MG, mixRing, ecdhInfo, outPk]
    :return: true / false
    """
    rvb = True 
    tmp = True 
    for i in range(0, len(rv.outPk)): 
        tmp = ver_range(rv.outPk[i].mask, rv.rangeSigs[i])
        rvb = rvb and tmp
    mgVerd = verify_rct_mg(rv.MG, rv.mixRing, rv.outPk)
    return rvb and mgVerd


def decode_rct(rv, sk, i):
    """
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1.1
    Uses the attached ecdh info to find the amounts represented by each output commitment
    must know the destination private key to find the correct amount, else will return a random number

    :param rv:
    :param sk:
    :param i:
    :return:
    """
    decodedTuple = ecdh_decode(rv.ecdhInfo[i], sk)
    mask = decodedTuple.mask
    amount = decodedTuple.amount
    C = rv.outPk[i].mask
    H = crypto.gen_H()
    Ctmp = crypto.point_add(crypto.scalarmult_base(mask), crypto.scalarmult(H, amount))
    if not crypto.point_eq(crypto.point_sub(C, Ctmp), crypto.identity()):
        logger.warning("warning, amount decoded incorrectly, will be unable to spend")
    return amount




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import unittest

import aiounittest

from monero_glue.misc import b58_mnr


class Base58Test(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(Base58Test, self).__init__(*args, **kwargs)

    def test_base(self):
        """
        Simple b58 encode test
        :return:
        """
        self.assertIsNotNone(b58_mnr.b58encode(b''))
        self.assertIsNotNone(b58_mnr.b58encode(b'1234567890'))
        self.assertIsNotNone(b58_mnr.b58encode(b'1234567890' * 50))
        self.assertEqual(
            b58_mnr.b58encode(
                b'123bec484c5d7f0246af520aab550452b5b6013733feabebd681c4a60d457b7fc12d5918e31d3c003da3c778592c07b398ad6f9'
                b'61a67082a75fd49394d51e69bbea9a6e386'),
            b'43tpGG9PKbwCpjRvNLn1jwXPpnacw2uVUcszAtgmDiVcZK4VgHwjJT9BJz1WGF9eMxSYASp8yNMkuLjeQfWqJn3CNWdWfzV')

        for i in range(10):
            inp = b'test01' * (7*i+i+1)
            enc = b58_mnr.b58encode(data_bin=inp)
            dec = b58_mnr.b58decode(data_bin=enc)
            self.assertEqual(inp, dec)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover




#https://tools.ietf.org/html/draft-josefsson-eddsa-ed25519-00#section-4
#Copyright (c) <YEAR>, <OWNER>
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are met:
#
#1. Redistributions of source code must retain the above copyright notice, this
   #list of conditions and the following disclaimer.
#2. Redistributions in binary form must reproduce the above copyright notice,
   #this list of conditions and the following disclaimer in the documentation
   #and/or other materials provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
#ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those
#of the authors and should not be interpreted as representing official policies,
#either expressed or implied, of the FreeBSD Project.

import hashlib
from . import ed25519

def sha512(s):
    return hashlib.sha512(s).digest()


# Base field Z_p
p = 2**255 - 19


def modp_inv(x):
    return pow(x, p-2, p)


# Curve constant
d = -121665 * modp_inv(121666) % p


# Group order
q = 2**252 + 27742317777372353535851937790883648493

# Points are represented as tuples (X, Y, Z, T) of extended coordinates,
# with x = X/Z, y = Y/Z, x*y = T/Z

def point_add(P, Q):
    A = (P[1]-P[0])*(Q[1]-Q[0]) % p
    B = (P[1]+P[0])*(Q[1]+Q[0]) % p
    C = 2 * P[3] * Q[3] * d % p
    D = 2 * P[2] * Q[2] % p
    E = B-A
    F = D-C
    G = D+C
    H = B+A
    return (E*F, G*H, F*G, E*H)


# Computes Q = s * Q
def point_mul(s, P):
    Q = (0, 1, 1, 0)  # Neutral element
    while s > 0:
        # Is there any bit-set predicate?
        if s & 1:
            Q = point_add(Q, P)
        P = point_add(P, P)
        s >>= 1
    return Q


def point_equal(P, Q):
    # x1 / z1 == x2 / z2  <==>  x1 * z2 == x2 * z1
    if (P[0] * Q[2] - Q[0] * P[2]) % p != 0:
        return False
    if (P[1] * Q[2] - Q[1] * P[2]) % p != 0:
        return False
    return True


# Square root of -1
modp_sqrt_m1 = pow(2, (p-1) // 4, p)


# Compute corresponding x coordinate, with low bit corresponding to sign,
# or return None on failure
def recover_x(y, sign):
    x2 = (y*y-1) * modp_inv(d*y*y+1)
    if x2 == 0:
        if sign:
            return None
        else:
            return 0

    # Compute square root of x2
    x = pow(x2, (p+3) // 8, p)
    if (x*x - x2) % p != 0:
        x = x * modp_sqrt_m1 % p
    if (x*x - x2) % p != 0:
        return None

    if (x & 1) != sign:
        x = p - x
    return x


# Base point
g_y = 4 * modp_inv(5) % p
g_x = recover_x(g_y, 0)
G = (g_x, g_y, 1, g_x * g_y % p)


def basepoint():
    return G


def point_compress(P):
    zinv = modp_inv(P[2])
    x = P[0] * zinv % p
    y = P[1] * zinv % p
    return ed25519.encodeint(y | ((x & 1) << 255))


def point_decompress(s):
    #if len(s) != 32:
        #raise Exception("Invalid input length for decompression")
    #y = int.from_bytes(s, "little")
    y = ed25519.decodeint(s)
    sign = y >> 255
    y &= (1 << 255) - 1

    x = recover_x(y, sign)
    if x is None:
        return None
    else:
        return x, y, 1, x*y % p




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import unittest

import aiounittest

from monero_glue.xmr import monero, ring_ct, crypto


class RingCtTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(RingCtTest, self).__init__(*args, **kwargs)

    def test_range_proof(self):
        proof = ring_ct.prove_range(0)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)

        proof = ring_ct.prove_range(0, mem_opt=False)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)

    def test_range_proof2(self):
        proof = ring_ct.prove_range(123456789)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)

        proof = ring_ct.prove_range(123456789, mem_opt=False)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)

        res = ring_ct.ver_range(crypto.point_add(proof[0], crypto.scalarmult_base(4)), proof[2])
        self.assertFalse(res)

    def test_range_proof3(self):
        proof = ring_ct.prove_range(123456789)
        rsig = proof[2]

        monero.recode_rangesig(rsig, encode=True)
        monero.recode_rangesig(rsig, encode=False)
        res = ring_ct.ver_range(proof[0], rsig)
        self.assertTrue(res)

    def test_range_proof_old(self):
        proof = ring_ct.prove_range(0, use_asnl=True, mem_opt=False)
        res = ring_ct.ver_range(proof[0], proof[2], use_asnl=True)
        self.assertTrue(res)

    def test_range_proof2_old(self):
        proof = ring_ct.prove_range(123456789, use_asnl=True, mem_opt=False)
        res = ring_ct.ver_range(proof[0], proof[2], use_asnl=True)
        self.assertTrue(res)
        res = ring_ct.ver_range(crypto.point_add(proof[0], crypto.scalarmult_base(4)), proof[2], use_asnl=True)
        self.assertFalse(res)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover


